"""View for defect object. Handle node in treeview and present forms to user when interacted with."""

from tkinter import TclError
import tkinter as tk
from pollenisatorgui.core.application.dialogs.ChildDialogFixView import ChildDialogFixView
from pollenisatorgui.core.application.dialogs.ChildDialogQuestion import ChildDialogQuestion
from pollenisatorgui.core.views.viewelement import ViewElement
from pollenisatorgui.core.models.defect import Defect
import pollenisatorgui.core.components.utils as utils
from pollenisatorgui.core.components.apiclient import APIClient
from PIL import ImageTk, Image
from shutil import which
from customtkinter import *
import os
import subprocess
from pollenisatorgui.core.application.dialogs.ChildDialogFixes import ChildDialogFixes

class DefectView(ViewElement):
    """View for defect object. Handle node in treeview and present forms to user when interacted with.
    Attributes:
        icon: icon name to show in treeview. Icon filename must be in icon directory.
    """

    icon = 'defect.png'

    def __init__(self, appTw, appViewFrame, mainApp, controller):
        """Constructor
        Args:
            appTw: a PollenisatorTreeview instance to put this view in
            appViewFrame: an view frame to build the forms in.
            mainApp: the Application instance
            controller: a CommandController for this view.
        """
        super().__init__(appTw, appViewFrame, mainApp, controller)
        self.easeForm = None
        self.impactForm = None
        self.riskForm = None

    def openInsertWindow(self, notes="", **kwargs):
        """
        Creates a tkinter form using Forms classes. This form aims to insert a new Defect
        Args:
            notes: default notes to be written in notes text input. Default is ""
            addButtons: boolean value indicating that insertion buttons should be visible. Default to True
        """
        self.form.kwargs["padx"] = 0
        settings = self.mainApp.settings
        settings.reloadSettings()
        apiclient = APIClient.getInstance()
        modelData = self.controller.getData()
        fullPanel = self.form.addFormPanel(grid=True)
        leftPanel = fullPanel.addFormPanel(column=0)
        rightPanel = fullPanel.addFormPanel(column=1, padx=1)
        s = leftPanel.addFormSearchBar("Search Defect", self.searchCallback, self.form, anchor="w")
        topPanel = leftPanel.addFormPanel(grid=True)
        topPanel.addFormLabel("Search Language",row=1, column=0)
        lang = topPanel.addFormCombo("Lang", apiclient.getLangList(), default=settings.db_settings.get("lang", "en"), width=100, row=1, column=1)
        topPanel.addFormLabel("Only defect for type",row=2, column=0)
        perimeter = topPanel.addFormStr("perimeter", default=settings.db_settings.get("pentest_type", ""), row=2, column=1)
        s.addOptionForm(lang, "lang")
        s.addOptionForm(perimeter, "perimeter")
        topPanel.addFormLabel("Results",row=3, column=0)
        result = topPanel.addFormCombo("Result", [""], default="", width=200, row=3, column=1)
        s.setResultForm(result)
        data_panel = leftPanel.addFormPanel(side="left", height=0, fill="x")
        topPanel = data_panel.addFormPanel(grid=True, side="top", pady=3)
        topPanel.addFormLabel("Title_lbl", text="Title")
        title = topPanel.addFormStr("Title", r".+", modelData.get("title", ""), placeholder="Title", column=1, width=400)
        s.addResultForm(title, "title")
        topPanel = data_panel.addFormPanel(grid=True, side="top", pady=3)
        topPanel.addFormLabel("Ease")
        self.easeForm = topPanel.addFormCombo(
            "Ease", Defect.getEases(), default=modelData.get("ease", ""), column=1, command=self.updateRiskBox, binds={"<<ComboboxSelected>>": self.updateRiskBox})
        s.addResultForm(self.easeForm, "ease")
        topPanel.addFormHelper("0: Trivial to exploit, no tool required\n1: Simple technics and public tools needed to exploit\n2: public vulnerability exploit requiring security skills and/or the development of simple tools.\n3: Use of non-public exploits requiring strong skills in security and/or the development of targeted tools", column=2)
        topPanel.addFormLabel("Impact", column=3)
        self.impactForm = topPanel.addFormCombo(
            "Impact", Defect.getImpacts(), default=modelData.get("impact", ""), command=self.updateRiskBox, column=4, binds={"<<ComboboxSelected>>": self.updateRiskBox})
        s.addResultForm(self.impactForm, "impact")
        topPanel.addFormHelper("0: No direct impact on system security\n1: Impact isolated on precise locations of pentested system security\n2: Impact restricted to a part of the system security.\n3: Global impact on the pentested system security.", column=5)
        topPanel.addFormLabel("Risk", column=6)
        self.riskForm = topPanel.addFormCombo(
            "Risk", Defect.getRisks(), modelData["risk"],  column=7)
        s.addResultForm(self.riskForm, "risk")
        topPanel.addFormHelper(
            "0: small risk that might be fixed\n1: moderate risk that need a planed fix\n2: major risk that need to be fixed quickly.\n3: critical risk that need an immediate fix or an immediate interruption.", column=8)
        topPanel = data_panel.addFormPanel(grid=True, side="top", pady=3)
        topPanel.addFormLabel("Redactor", row=1)
        topPanel.addFormCombo("Redactor", self.mainApp.settings.getPentesters()+["N/A"], "N/A", row=1, column=1)
        topPanel.addFormLabel("Language", row=1, column=2)
        lang = topPanel.addFormStr("Language", "", modelData.get("language","en"), row=1, column=3)
        s.addResultForm(lang, "language")
        chklistPanel = data_panel.addFormPanel( side="top", pady=3, anchor="center")
        defectTypes = settings.getPentestTypes()
        if defectTypes is not None:
            defectTypes = defectTypes.get(settings.getPentestType(), [])
            if len(defectTypes) == 0:
                defectTypes = ["N/A"]
        else:
            defectTypes = ["N/A"]
        checklist = chklistPanel.addFormChecklist("Type", defectTypes, ["N/A"], anchor="center",fill="x")
        s.addResultForm(checklist, "type")
        topPanel = data_panel.addFormPanel(side="top", pady=3)
        settings = self.mainApp.settings
        topPanel.addFormLabel("Synthesis", side="top")
        synthesis = topPanel.addFormText("Synthesis", r"", modelData.get("synthesis", "Synthesis"), state="readonly" if self.controller.isAssigned() else "", side="left", height=200)
        s.addResultForm(synthesis, "synthesis")
        if not self.controller.isAssigned():
            topPanel = data_panel.addFormPanel(side="top", pady=3)
            topPanel.addFormLabel("Fixes", side="top")
            fixesPane = topPanel.addFormPanel(side=tk.TOP, fill=tk.X)
            values = []
            for fix in modelData.get("fixes", []):
                values.append((fix["title"], fix["execution"], fix["gain"], fix["synthesis"], fix["description"]))
            fixesPane.addFormButton("Add fix", self.addFix, side=tk.RIGHT)
            self.fix_treevw = fixesPane.addFormTreevw("Fixes", ("Title", "Execution", "Gain"), values, height=3, max_height=5, anchor=tk.CENTER, 
                                                    side=tk.RIGHT, auto_size_columns=False,
                                                    doubleClickBinds=[self.onFixDoubleClick, self.onFixDoubleClick, self.onFixDoubleClick])
            s.addResultForm(self.fix_treevw, "fixes", fill_callback=self.fillFixes)
            topPanel = rightPanel.addFormPanel()
            desc = topPanel.addFormMarkdown("Description", r"", modelData.get("description", "Description"), side="top", fill=tk.BOTH)
            s.addResultForm(desc, "description")
            
        else:
            topPanel.addFormHidden("Description", modelData.get("description", ""))
            notesPanel = rightPanel.addFormPanel()
            notesPanel.addFormLabel("Notes", side="top")
            notesPanel.addFormText("Notes", r"", modelData.get("notes", notes), None, side="top")
            self.form.addFormHidden("fixes", modelData.get("fixes", {}))
        self.form.addFormHidden("target_id", modelData["target_id"])
        self.form.addFormHidden("target_type", modelData["target_type"])
        if kwargs.get("addButtons", True):
            self.completeInsertWindow()
        else:
            self.showForm()
        self.updateRiskBox()

    def fillFixes(self, fix_form, fixes):
        fix_form.reset()
        values = []
        for fix in fixes:
            values.append((fix["title"], fix["execution"], fix["gain"], fix["synthesis"], fix["description"]))
        fix_form.recurse_insert(values)

    def openModifyWindow(self, **kwargs):
        """
        Creates a tkinter form using Forms classes.
        This form aims to update or delete an existing Defect
        Args:
            addButtons: boolean value indicating that insertion buttons should be visible. Default to True
        """
        self.form.clear()
        self.form.kwargs["fill"] = "both"
        self.form.kwargs["expand"] = True
        self.form.kwargs["padx"] = 0
        modelData = self.controller.getData()
        settings = self.mainApp.settings
        settings.reloadSettings()
        self.delete_image = CTkImage(Image.open(utils.getIcon('delete.png')))
        self.edit_image = CTkImage(Image.open(utils.getIcon('stylo.png')))
        globalPanel = self.form.addFormPanel(side=tk.TOP, fill=tk.BOTH)
        leftPanel = globalPanel.addFormPanel(grid=True, side=tk.LEFT, fill=tk.Y, anchor="center")
        rightPanel = globalPanel.addFormPanel(grid=False, side=tk.RIGHT, fill=tk.BOTH, padx=5)
        row = 0
        if modelData.get("target_id", "") != "":
            leftPanel.addFormLabel("Target", row=row, column=0)
            target = self.controller.getTargetRepr()
            leftPanel.addFormStr(
                "Target", '', target, None, column=1, row=row, state="readonly")
            row += 1
        if not self.controller.model.isTemplate:
            if modelData["proofs"]:
                for i, proof in enumerate(modelData["proofs"]):
                    proof_local_path = self.controller.getProofWithName(proof)
                    if proof_local_path is not None:
                        modelData["description"] = modelData["description"].replace(f"[{proof}]({proof})", f"[{proof}]({proof_local_path})")
        if not self.controller.isAssigned():
            # if not self.controller.model.isTemplate:
            #     topPanel.addFormSearchBar("Search Defect", APIClient.getInstance().searchDefect, globalPanel, row=row, column=1, autofocus=False)
            #     row += 1
            titlePanel = leftPanel.addFormPanel(grid=True,  row=row, column=0, columnspan=2)
            titlePanel.addFormLabel("Title_lbl", text="Title", row=0, column=0)
            titlePanel.addFormStr(
                "Title", ".+", modelData["title"],  width=400, row=0, column=1)
            row += 1
            easePanel = leftPanel.addFormPanel(grid=True,  row=row, column=0, columnspan=2)
            easePanel.addFormLabel("Ease", row=0)
            self.easeForm = easePanel.addFormCombo(
                "Ease", Defect.getEases(), modelData["ease"], command=self.updateRiskBox, row=0, column=1,
                 binds={"<<ComboboxSelected>>": self.updateRiskBox})
            easePanel.addFormHelper("0: Trivial to exploit, no tool required\n1: Simple technics and public tools needed to exploit\n2: public vulnerability exploit requiring security skills and/or the development of simple tools.\n3: Use of non-public exploits requiring strong skills in security and/or the development of targeted tools", 
                                    row=0, column=2)
            easePanel.addFormLabel("Impact", row=1, column=0)
            self.impactForm = easePanel.addFormCombo(
                "Impact", Defect.getImpacts(), modelData["impact"], command=self.updateRiskBox,
                 row=1, column=1, binds={"<<ComboboxSelected>>": self.updateRiskBox})
            easePanel.addFormHelper("0: No direct impact on system security\n1: Impact isolated on precise locations of pentested system security\n2: Impact restricted to a part of the system security.\n3: Global impact on the pentested system security.",
                                     row=1, column=2)
            easePanel.addFormLabel("Risk", row=2, column=0)
            self.riskForm = easePanel.addFormCombo(
                "Risk", Defect.getRisks(), modelData["risk"],  row=2, column=1)
            easePanel.addFormHelper(
                "0: small risk that might be fixed\n1: moderate risk that need a planed fix\n2: major risk that need to be fixed quickly.\n3: critical risk that need an immediate fix or an immediate interruption.",
                 row=2, column=2)
            row += 1
            chklistPanel = leftPanel.addFormPanel(grid=True,  row=row, column=0, columnspan=2)
            defect_types = settings.getPentestTypes().get(settings.getPentestType(), [])
            for savedType in modelData["type"]:
                if savedType.strip() not in defect_types:
                    defect_types.insert(0, savedType)
            chklistPanel.addFormChecklist("Type", defect_types, modelData["type"])
            row += 1
            otherPanel = leftPanel.addFormPanel(grid = True,  row=row, column=0, columnspan=2)
            pos_row = 0
            if not self.controller.model.isTemplate:
                otherPanel.addFormLabel("Redactor", row=pos_row)
                otherPanel.addFormCombo("Redactor", list(set(self.mainApp.settings.getPentesters()+["N/A"]+[modelData["redactor"]])), modelData["redactor"], row=pos_row, column=1)
                pos_row += 1
            otherPanel.addFormLabel("Language", row=pos_row, column=0)
            otherPanel.addFormStr("Language", "", modelData["language"], row=pos_row, column=1)
            row += 1
            otherPanel = leftPanel.addFormPanel( row=row, column=0, columnspan=2)
            otherPanel.addFormLabel("Synthesis", side=tk.TOP)
            otherPanel.addFormText("Synthesis", r"", modelData.get("synthesis","Synthesis"), state="readonly" if self.controller.isAssigned() else "",  side="top", fill=tk.NONE)
            row += 1
            fixesPane = leftPanel.addFormPanel(row=row, column=0, columnspan=2)
            fixesPane.addFormLabel("Fixes", side=tk.TOP)
            values = []
            for fix in modelData["fixes"]:
                values.append((fix["title"], fix["execution"], fix["gain"], fix["synthesis"], fix["description"]))
            fixesPane.addFormButton("Add fix", self.addFix, side=tk.RIGHT,width=10)
            self.fix_treevw = fixesPane.addFormTreevw("Fixes", ("Title", "Execution", "Gain"), values, padx=1, height=3, max_height=5, anchor=tk.CENTER, 
                                                    side=tk.RIGHT, auto_size_columns=False,
                                                    doubleClickBinds=[self.onFixDoubleClick, self.onFixDoubleClick, self.onFixDoubleClick])
        
            self.description_form = rightPanel.addFormMarkdown("Description", r"", modelData.get("description", "Description"),  style_change=True, just_editor=True, fill=tk.BOTH, expand=True)
        else:
            globalPanel.addFormHidden("Title", modelData.get("title", ""))
            globalPanel.addFormHidden("Ease", modelData.get("ease", ""))
            globalPanel.addFormHidden("Impact", modelData.get("impact", ""))
            globalPanel.addFormHidden("Risk", modelData.get("risk", ""))
            types = modelData.get("type", [])
            type_dict = dict()
            for type in types:
                type_dict[type] = 1
            globalPanel.addFormHidden("Type", type_dict)
            globalPanel.addFormHidden("Language", modelData.get("language", ""))
            globalPanel.addFormHidden("Synthesis", modelData.get("synthesis", ""))
            globalPanel.addFormHidden("Description", modelData.get("description", ""))
            notesPanel = globalPanel.addFormPanel()
            notesPanel.addFormLabel("Notes", side="top")
            notesPanel.addFormText(
                "Notes", r"", modelData["notes"], None, side="top", height=40)
            
            fixesPane = self.form.addFormPanel(side=tk.TOP, anchor=tk.CENTER, fill=tk.X)
            values = []
            for fix in modelData["fixes"]:
                values.append((fix["title"], fix["execution"], fix["gain"], fix["synthesis"], fix["description"]))
            fixesPane.addFormButton("Add fix", self.addFix, side=tk.RIGHT, width=10)
            self.fix_treevw = fixesPane.addFormTreevw("Fixes", ("Title", "Execution", "Gain"), values, height=3, max_height=5, anchor=tk.CENTER, 
                                                    side=tk.RIGHT, auto_size_columns=False,
                                                    doubleClickBinds=[self.onFixDoubleClick, self.onFixDoubleClick, self.onFixDoubleClick])
            
        #self.formFixes = globalPanel.addFormHidden("Fixes", modelData["fixes"])
        if not self.controller.model.isTemplate:
            actionsPan = self.form.addFormPanel(side=tk.TOP, anchor=tk.E)
            #actionsPan.addFormButton("Edit fixes", self.openFixesWindow, side=tk.RIGHT, image=self.edit_image)
            actionsPan.addFormButton("Create defect template from this", self.saveAsDefectTemplate,  image=self.edit_image, side=tk.RIGHT )
        if kwargs.get("addButtons", True):
            self.completeModifyWindow(addTags=False)
        else:
            self.showForm()

    def addFix(self, _event=None):
        dialog = ChildDialogFixView(None)
        dialog.app.wait_window(dialog.app)
        if dialog.rvalue is None:
            return
        self.fix_treevw.addItem("", "end", dialog.rvalue["title"],  text=dialog.rvalue["title"], values=(dialog.rvalue["execution"], dialog.rvalue["gain"],  dialog.rvalue["synthesis"], dialog.rvalue["description"]))

    def onFixDoubleClick(self, event):
        selected = self.fix_treevw.selection()
        if not selected:
            return
        selected = selected[0]
        item = self.fix_treevw.item(selected)
        title = item["text"]
        fix = {"title":title, "gain": item["values"][1], "execution": item["values"][0], "synthesis": item["values"][2], "description": item["values"][3]}
        dialog = ChildDialogFixView(None, fix)
        dialog.app.wait_window(dialog.app)
        if dialog.rvalue is None:
            return
        self.fix_treevw.item(selected, text=dialog.rvalue["title"], values=(dialog.rvalue["execution"], dialog.rvalue["gain"],  dialog.rvalue["synthesis"], dialog.rvalue["description"]))


    def insert(self, _event=None):
        """
        Entry point to the model doInsert function.

        Args:
            _event: automatically filled if called by an event. Not used
        Returns:
            * a boolean to shwo success or failure
            * an empty message on success, an error message on failure
        """
        res, msg = super().insert(_event=None)
        if res:
            apiclient = APIClient.getInstance()
            results, msg = apiclient.searchDefect(self.controller.model.title)#, check_api=False
            if results is not None and len(results) == 0:
                dialog = ChildDialogQuestion(self.mainApp, "Create defect template", "This defect seems new. Do you want to create a defect template with this defect?")
                self.mainApp.wait_window(dialog.app)
                if dialog.rvalue == "Yes":
                    self.saveAsDefectTemplate()
        return res, msg

    def openMultiModifyWindow(self, **kwargs):
        self.form.clear()
        settings = self.mainApp.settings
        settings.reloadSettings()
        apiclient = APIClient.getInstance()
        results, msg = apiclient.searchDefect("")#, check_api=False
        default_values = {}
        formFilter = self.form.addFormPanel(grid=True)
        lbl_filter_title = formFilter.addFormLabel("Filters")
        self.str_filter_title = formFilter.addFormStr("Title", "", placeholder_text="title", row=0, column=1, binds={"<Key-Return>":  self.filter})
        formFilter.addFormLabel("Risk", row=0, column=2)
        risks = set([result["risk"] for result in results]+[""])
        self.box_filter_risk = formFilter.addFormCombo("Risk", risks, "", command=self.filter, width=100, row=0, column=3)
        formFilter.addFormLabel("Perimeter", row=0, column=4)
        default_perimeter = settings.getPentestType()
        perimeters = set()
        perimeters.add(default_perimeter)
        perimeters.add("")
        for result in results:
            for perimeter in result.get("perimeter", "").split(","):
                perimeters.add(perimeter)
        self.box_filter_perimeter = formFilter.addFormCombo("Perimeter", perimeters, default_perimeter , command=self.filter, width=100, row=0, column=5)
        formFilter.addFormLabel("Lang", row=0, column=6)
        langs = set([result["language"] for result in results]+[""])
        default_lang = settings.db_settings.get("lang", "en")
        self.box_filter_lang = formFilter.addFormCombo("Lang", langs, default_lang, command=self.filter, width=100, row=0, column=7)
        formFilterSecond = self.form.addFormPanel(grid=True)
        lbl_filter_title = formFilterSecond.addFormLabel("Show only results from")
        self.box_filter_source = formFilterSecond.addFormCombo("Source", ["All", "Local Templates", "API"], "All", command=self.filter, row=0, column=1)
        formTreevw = self.form.addFormPanel()
        if results is not None:
            for result in results:
                if result is not None:
                    default_values[str(result.get("_id", result.get("id")))] = (result["title"], result["risk"], result["language"], result.get("perimeter", settings.getPentestType()), result["source"])
            self.browse_top_treevw = formTreevw.addFormTreevw("Defects", ("Title", "Risk", "Lang", "Perimeter", "Source"),
                                default_values, side="top", fill="both", width=500, height=8, status="readonly", 
                                binds={"<Double-Button-1>":self.doubleClickDefectView, "<Delete>":self.deleteDefectTemplate})
            
        panel_action = self.form.addFormPanel()
        self.import_image = CTkImage(Image.open(utils.getIcon("import.png")))
        self.export_image = CTkImage(Image.open(utils.getIcon("download.png")))
        panel_action.addFormButton("Import", self.importDefectTemplates, image=self.import_image)
        panel_action.addFormButton("Export", self.exportDefectTemplates, image=self.export_image)
        if kwargs.get("addButtons", True):
            self.completeModifyWindow()
        else:
            self.showForm()
        self.filter()

    def openMultiInsertWindow(self, **kwargs):
        """
        Creates a tkinter form using Forms classes. This form aims to insert many Defects
        Args:
            addButtons: boolean value indicating that insertion buttons should be visible. Default to True
        """
        settings = self.mainApp.settings
        settings.reloadSettings()
        apiclient = APIClient.getInstance()
        results, msg = apiclient.searchDefect("")
        default_values = {}
        formFilter = self.form.addFormPanel(grid=True)
        lbl_filter_title = formFilter.addFormLabel("Filters")
        self.str_filter_title = formFilter.addFormStr("Title", "", placeholder_text="title", row=0, column=1, binds={"<Key-Return>":  self.filter})
        formFilter.addFormLabel("Risk", row=0, column=2)
        risks = set([result["risk"] for result in results]+[""])
        self.box_filter_risk = formFilter.addFormCombo("Risk", risks, "", command=self.filter, width=100, row=0, column=3)
        formFilter.addFormLabel("Perimeter", row=0, column=4)
        default_perimeter = settings.getPentestType()
        perimeters = set()
        perimeters.add(default_perimeter)
        perimeters.add("")
        for result in results:
            for perimeter in result["perimeter"].split(","):
                perimeters.add(perimeter)
        self.box_filter_perimeter = formFilter.addFormCombo("Perimeter", perimeters, default_perimeter , command=self.filter, width=100, row=0, column=5)
        formFilter.addFormLabel("Lang", row=0, column=6)
        langs = set([result["language"] for result in results]+[""])
        default_lang = settings.db_settings.get("lang", "en")
        self.box_filter_lang = formFilter.addFormCombo("Lang", langs, default_lang, command=self.filter, width=100, row=0, column=7)
        formFilterSecond = self.form.addFormPanel(grid=True)
        lbl_filter_title = formFilterSecond.addFormLabel("Show only results from")
        self.box_filter_source = formFilterSecond.addFormCombo("Source", ["All", "Local Templates", "API"], "All", command=self.filter, row=0, column=1)
        formTreevw = self.form.addFormPanel()
        if results is not None:
            for result in results:
                if result is not None:
                    default_values[str(result.get("_id", result.get("id")))]  = (result["title"], result["risk"], result["language"], result["perimeter"], result["source"])
            self.browse_top_treevw = formTreevw.addFormTreevw("Defects", ("Title", "Risk", "Lang", "Perimeter", "Source"),
                                default_values, side="top", fill="both", width=500, height=8, status="readonly", 
                                binds={"<Double-Button-1>":self.doubleClickDefectView, "<Delete>":self.deleteDefectTemplate})
            
        self.buttonUpImage = CTkImage(Image.open(utils.getIcon('up-arrow.png')))
        self.buttonDownImage = CTkImage(Image.open(utils.getIcon('down-arrow.png')))
        # use self.buttonPhoto
        buttonPan = self.form.addFormPanel(side="top", anchor="center", fill="none")
        btn_down = buttonPan.addFormButton("Add to report", self.moveDownMultiTreeview, side="left", anchor="center", image=self.buttonDownImage)
        btn_down = buttonPan.addFormButton("Remove from report", self.moveUpMultiTreeview, side="right", anchor="center", image=self.buttonUpImage)
        default_values = {}
        self.browse_down_treevw = self.form.addFormTreevw("Defects", ("Title", "Risk"),
                            default_values, side="bottom", fill="both", width=500, height=8, status="readonly")
        if kwargs.get("addButtons", True):
            self.completeInsertWindow()
        else:
            self.showForm()
        self.filter()

    def filter(self, event=None):
        title = self.str_filter_title.getValue()
        lang = self.box_filter_lang.getValue()
        perimeter = self.box_filter_perimeter.getValue()
        risk = self.box_filter_risk.getValue()
        source = self.box_filter_source.getValue()
        if source == "Local Templates":
            source = "local"
        elif source == "All":
            source = True
        else:
            source = source.lower()
        
        self.browse_top_treevw.filter(title, risk, lang, perimeter, source)

    
    def searchCallback(self, searchreq, **options):
        defects_obj, defects_errors = APIClient.getInstance().searchDefect(searchreq, **options)
        if defects_obj:
            for i, defect in enumerate(defects_obj):
                defects_obj[i]["TITLE"] = defect["title"]
        return defects_obj, defects_errors

    def deleteDefectTemplate(self, event):
        apiclient = APIClient.getInstance()
        sel = self.browse_top_treevw.treevw.selection()
        if len(sel) > 1:
            answer = tk.messagebox.askyesno(
                        "Defect template deletion warning", f"{len(sel)} defects will be deleted from the defect templates database. Are you sure ?")
        for selected in sel:
            item = self.browse_top_treevw.treevw.item(selected)
            if item["text"].strip() != "":
                defect_m = self.findDefectTemplateByTitle(item["text"].strip())
                if isinstance(defect_m, Defect):
                    if len(sel) == 1:
                        answer = tk.messagebox.askyesno(
                            "Defect template deletion warning", f"{defect_m.title} will be deleted from the defect templates database. Are you sure ?")
                        if not answer:
                            return
                    apiclient.deleteDefectTemplate(defect_m.getId())
        self.browse_top_treevw.deleteItem()

    def openInChildDialog(self, defect_model, isTemplate=True):
        from pollenisatorgui.core.application.dialogs.ChildDialogDefectView import ChildDialogDefectView
        defect_model.isTemplate = isTemplate
        if defect_model.isTemplate:
            title = "Edit a security defect template"
        else:
            title = "Edit a security defect"
        dialog = ChildDialogDefectView(self.mainApp, title, self.mainApp.settings, defect_model)
        self.mainApp.wait_window(dialog.app)
        return dialog.rvalue

    def findDefectTemplateByTitle(self, title, multi=False):
        apiclient = APIClient.getInstance()
        defects_matching, msg = apiclient.searchDefect(title)# check_api=False
        if defects_matching is not None:
            if len(defects_matching) >= 1 and not multi:
                return Defect(defects_matching[0])
            else:
                return defects_matching
            

    def doubleClickDefectView(self, event):
        item = self.browse_top_treevw.treevw.identify("item", event.x, event.y)
        if item is None or item == '':
            return
        title = self.browse_top_treevw.treevw.item(item)["text"]
        defects_matching = self.findDefectTemplateByTitle(title)
        self.openInChildDialog(defects_matching)

    def moveDownMultiTreeview(self, _event=None):
        for iid in self.browse_top_treevw.selection():
            item = self.browse_top_treevw.item(iid)
            self.browse_down_treevw.addItem("","end", iid, text=item["text"], values=item["values"])
        self.browse_top_treevw.deleteItem()

    def moveUpMultiTreeview(self, _event=None):
        for iid in self.browse_down_treevw.selection():
            item = self.browse_down_treevw.item(iid)
            self.browse_top_treevw.addItem("","end", iid, text=item["text"], values=item["values"])
        self.browse_down_treevw.deleteItem()

    

    def openFixesWindow(self, _event=None):
        dialog = ChildDialogFixes(None, self)
        dialog.app.wait_window(dialog.app)
        if dialog.rvalue is None:
            return
        self.formFixes.setValue(dialog.rvalue)
        self.controller.update_fixes(dialog.rvalue)

    def updateRiskBox(self, _event=None):
        """Callback when ease or impact is modified.
        Calculate new resulting risk value
        Args
            _event: mandatory but not used
        """
        ease = self.easeForm.getValue()
        impact = self.impactForm.getValue()
        risk = Defect.getRisk(ease, impact)
        self.riskForm.setValue(risk)

    def viewProof(self, _event, obj):
        """Callback when view proof is clicked.
        Download and display the file 
        Args
            _event: mandatory but not used
            obj: the clicked index proof
        """
        proof_local_path = self.controller.getProof(obj)
        if proof_local_path is not None:
            if os.path.isfile(proof_local_path):
                res = utils.openPathForUser(proof_local_path)
                if not res:
                    tk.messagebox.showerror("Could not open", "Failed to open this file.")
                    proof_local_path = None
                    return
        if proof_local_path is None:
            tk.messagebox.showerror(
                "Download failed", "the file does not exist on sftp server")

    def deleteProof(self, _event, obj):
        """Callback when delete proof is clicked.
        remove remote proof and update window
        Args
            _event: mandatory but not used
            obj: the clicked index proof
        """
        self.controller.deleteProof(obj)
        self.form.clear()
        for widget in self.appliViewFrame.winfo_children():
            widget.destroy()
        self.openModifyWindow()

    def addAProof(self, _event=None):
        """Callback when add proof is clicked.
        Add proof and update window
        Args
            _event: mandatory but not used
        """
        values = self.formFile.getValue()
        for val in values:
            self.controller.addAProof(val)
        self.form.clear()
        for widget in self.appliViewFrame.winfo_children():
            widget.destroy()
        self.openModifyWindow()

    def beforeDelete(self, iid=None):
        """Called before defect deletion.
        Will attempt to remove this defect from global defect table.
        Args:
            iid: the mongo ID of the deleted defect
        """
        if iid is None:
            if self.controller is not None:
                iid = self.controller.getDbId()
        if iid is not None:
            for module in self.mainApp.modules:
                if callable(getattr(module["object"], "removeItem", None)):
                    module["object"].removeItem(iid)

    def addInTreeview(self, parentNode=None, **kwargs):
        """Add this view in treeview. Also stores infos in application treeview.
        Args:
            parentNode: if None, will calculate the parent. If setted, forces the node to be inserted inside given parentNode.
            _addChildren: not used here
        """
        self.appliTw.views[str(self.controller.getDbId())] = {"view": self}
        if not self.controller.isAssigned():
            # Unassigned defect are loaded on the report tab
            return
        if self.controller.model is None:
            return
        
        
        if parentNode is None:
            parentNode = DefectView.DbToTreeviewListId(
                self.controller.getParentId())
            nodeText = str(self.controller.getModelRepr())
        elif parentNode == '':
            nodeText = self.controller.getDetailedString()
        else:
            parentNode = DefectView.DbToTreeviewListId(parentNode)
            nodeText = str(self.controller.getModelRepr())
        try:
            parentNode = self.appliTw.insert(
                self.controller.getParentId(), 0, parentNode, text="Defects", image=self.getIcon())
        except TclError:
            pass
        try:
            self.appliTw.insert(parentNode, "end", str(self.controller.getDbId()),
                                text=nodeText, tags=self.controller.getTags(), image=self.getIcon())
        except TclError:
            pass
    
        if "hidden" in self.controller.getTags():
            self.hide("tags")
        if self.mainApp.settings.is_checklist_view():
            self.hide("checklist_view")

    @classmethod
    def DbToTreeviewListId(cls, parent_db_id):
        """Converts a mongo Id to a unique string identifying a list of defects given its parent
        Args:
            parent_db_id: the parent node mongo ID
        Returns:
            A string that should be unique to describe the parent list of defect node
        """
        return str(parent_db_id)+"|Defects"

    @classmethod
    def treeviewListIdToDb(cls, treeviewId):
        """Extract from the unique string identifying a list of defects the parent db ID
        Args:
            treeviewId: the treeview node id of a list of defects node
        Returns:
            the parent object mongo id as string
        """
        return str(treeviewId).split("|")[0]

    def multi_insert(self):
        values = self.browse_down_treevw.getValue()
        errors = []
        msg = ""
        for title in values:
            results, msg = APIClient.getInstance().searchDefect(title)
            if results is not None and results:
                result = results[0]
                d_o = Defect()
                if isinstance(result.get("type"), str):
                    types = result["type"].split(",")
                elif isinstance(result.get("type"), list):
                    types = result.get("type")
                else:
                    tk.messagebox.showerror("Multi insert error", f"Invalid defect result for : {title}. Wrong type : {result.get('type')}")
                    return False
                d_o.initialize("", "", result["title"], result["synthesis"], result["description"],
                            result["ease"], result["impact"], result["risk"], "N/A", types, result["language"], "", None, result["fixes"])
                d_o.addInDb()
            else:
                errors.append(title)
        if errors:
            msg = "Could not find the following defects in the knowledge db : \n"
            msg +=", ".join(errors)
            #    tk.messagebox.showerror("Could not search defect from knowledge db", msg)
            return False, msg
        return True, "Success"

    def insertReceived(self):
        """Called when a defect insertion is received by notification.
        Insert the node in treeview.
        Also insert it in global report of defect
        """
        if self.controller.model is None:
            return
        if self.controller.isAssigned():
            super().insertReceived()
        else:
            for module in self.mainApp.modules:
                if callable(getattr(module["object"], "addDefect", None)):
                    module["object"].addDefect(self.controller.model)
    
    def updateReceived(self, obj=None, old_obj=None):
        """Called when a defect update is received by notification.
        Update the defect node and the report defect table.
        """
        if self.controller.model is None:
            return
        if not self.controller.isAssigned():
            for module in self.mainApp.modules:
                if callable(getattr(module["object"], "updateDefectInTreevw", None)):
                    module["object"].updateDefectInTreevw(self.controller.model)
        super().updateReceived()

    def saveAsDefectTemplate(self, _event=None):
        settings = self.mainApp.settings
        settings.reloadSettings()
        perimeter = settings.getPentestType()
        self.controller.model.perimeter = perimeter 
        res = self.controller.model.addInDefectTemplates()
        defect_m = self.findDefectTemplateByTitle(self.controller.model.title)
        self.openInChildDialog(defect_m)

    def importDefectTemplates(self, _event=None):
        filename = ""
        f = tk.filedialog.askopenfilename(defaultextension=".json")
        if f is None:  # asksaveasfile return `None` if dialog closed with "cancel".
            return
        filename = str(f)
        try:
            apiclient = APIClient.getInstance()
            success = apiclient.importDefectTemplates(filename)
        except IOError:
            tk.messagebox.showerror(
                "Import defects templates", "Import failed. "+str(filename)+" was not found or is not a file.")
            return False
        if not success:
            tk.messagebox.showerror("Defects templates import", "Defects templatest failed")
        else:
            tk.messagebox.showinfo("Defects templates import", "Defects templates completed")
        self.openMultiModifyWindow(False)
        return success
    
    def exportDefectTemplates(self, _event=None):
        apiclient = APIClient.getInstance()
        res = apiclient.exportDefectTemplates(self.mainApp)
        if res is None:
            return
        else:
            res, msg = res # unpack tuple
        if res:
            tk.messagebox.showinfo(
                "Export Defect templates", "Export completed in "+str(msg))
        else:
            tk.messagebox.showinfo(msg)

    


"""This class pop a defect view form in a subdialog"""

import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
import pollenisatorgui.core.components.utils as utils
from pollenisatorgui.core.application.scrollableframexplateform import ScrollableFrameXPlateform
class ChildDialogView:
    """
    Open a child dialog of a tkinter application to answer a question.
    """
    def __init__(self, parent, title, scrollable=False):
        """
        Open a child dialog of a tkinter application to choose autoscan settings.

        Args:
            parent: the tkinter parent view to use for this window construction.
            model : A Defect model object to load default values. None to have empty fields, default is None.
        """
        self.app = CTkToplevel(parent)
        self.app.attributes("-type", "dialog")
        self.app.title(title)
        self.parent = parent
        self.app.resizable(True, True)
        self.app.bind("<Escape>", self.cancel)
        #self.app.geometry("1000x1000")
        if scrollable:
            self.appFrame = ScrollableFrameXPlateform(self.app)
        else:
            self.appFrame = CTkFrame(self.app)
        self.appFrame.columnconfigure(0, weight=1)
        self.appFrame.rowconfigure(0, weight=1)
        self.rvalue = None
        
    def completeDialogView(self, **kwargs):
        if kwargs.get("addButtons", True):
            ok_button = CTkButton(self.appFrame, text=kwargs.get("text_ok", "OK"))
            ok_button.pack(side="right", padx=5, pady=10)
            ok_button.bind('<Button-1>', self.okCallback)
            cancel_button = CTkButton(self.appFrame, text="Cancel", fg_color=utils.getBackgroundColor(), text_color=utils.getTextColor(),
                               border_width=1, border_color="firebrick1", hover_color="tomato")
            cancel_button.pack(side="right", padx=5, pady=10, ipadx=3)
            cancel_button.bind('<Button-1>', self.cancel)
        self.appFrame.pack(fill=tk.BOTH, ipady=10, ipadx=10, expand=True)
        self.app.update()
        # self.appFrame.pack(fill=tk.X, ipady=10, ipadx=3, expand=True) this break the canvas drawing with scrollbar
        try:
            self.app.wait_visibility()
            self.app.transient(self.parent)
            self.app.grab_set()
            self.app.focus_force()
            self.app.lift()
        except tk.TclError:
            pass

    def cancel(self, _event=None):
        """called when canceling the window.
        Close the window and set rvalue to False
        Args:
            _event: Not used but mandatory"""
        self.rvalue = False
        self.app.destroy()

    def okCallback(self, _event=None):  
        """called when pressing the validating button
        Close the window if the form is valid.
        Set rvalue to True and perform the defect update/insert if validated.
        To be overriden
        Args:
            _event: Not used but mandatory"""
        
        pass




from pollenisatorgui.core.components.apiclient import APIClient
import pollenisatorgui.core.components.utils as utils
import tempfile
import os
import shutil

def main(apiclient, appli, **kwargs):
    cme_path = utils.which_expand_alias("cme")
    if not utils.which_expand_alias("cme"):
        return False, "binary 'cme' is not in the PATH."
    APIClient.setInstance(apiclient)
    unk_users = apiclient.find("users", {"type":"user", "password":""})
    users_to_test = set()
    domains = set()
    for user in unk_users:
        users_to_test.add((user["domain"], user["username"]))
        domains.add(user["domain"])
    dcs = {}
    exec = 0
    for domain in domains:
        dc_info = apiclient.find("computers", {"type":"computer", "domain":domain, "infos.is_dc":True}, False)
        if dc_info is None:
            continue
        temp_folder = tempfile.gettempdir() 
        file_name = os.path.join(temp_folder, "users_"+str(domain)+".txt")
        users = "\n".join(x[1] for x in users_to_test if x[0] == domain)
        if users.strip() != "":
            with open(file_name, "w") as f:
                f.write(users+"\n")
            exec += 1
            appli.launch_in_terminal(kwargs.get("default_target", None), "cme bruteforce", f"{cme_path} smb {dc_info['ip']} -u {file_name} -p {file_name} -d {domain} --no-bruteforce --continue-on-success")
    return True, f"Launched {exec} cmes"



from collections import OrderedDict
from customtkinter import *
import tkinter.ttk as ttk


class Paginable(CTkFrame):

    def __init__(self, parent, callback_insert, callback_resetview, callback_get_value, callback_update_content_view, **kwargs) -> None:
        self.maxPerPage = kwargs.get("maxPerPage", 10)
        if "maxPerPage" in kwargs:
            del kwargs["maxPerPage"]
        super().__init__(parent, **kwargs)
        self.disablePagination = self.maxPerPage < 1
        self.currentPage = 0
        self.lastPage = 0
        self.pagePanel = None
        self.infos = OrderedDict()
        self._save_infos = None
        self.contentview = CTkFrame(self, height=0)
        self.contentview.grid(row=0, column=0, sticky="nsew")
        self.contentview.columnconfigure(0, weight=1)
        self.contentview.rowconfigure(0, weight=1)
        self.pagePanel = CTkFrame(self, height=0)
        self.pagePanel.grid(row=1, column=0, sticky="sew")
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self.rowconfigure(2, weight=0)
        self.callback_insert = callback_insert
        self.callback_get_value = callback_get_value
        self.callback_resetview = callback_resetview
        self.update_contentview = callback_update_content_view

    def addPaginatedInfo(self, info, auto_update_pagination=True):
        if "iid" not in info:
            raise ValueError("iid key not found in info")
        self.infos[str(info["iid"])] = info
        if self.disablePagination:
            return self.callback_insert([info])
        nbLig = len(self.infos)
        prevLastPage = self.lastPage
        self.lastPage = int(nbLig / self.maxPerPage)
        if prevLastPage != self.lastPage and self.currentPage +5 < self.lastPage and auto_update_pagination:
            self.setPaginationPanel()
        if int(nbLig % self.maxPerPage) == 0:
            self.lastPage -= 1
        res = None
        if self.currentPage == self.lastPage:
            shown = nbLig % self.maxPerPage
            if shown < self.maxPerPage:
                res = self.callback_insert([info])
        return res  

    def getShownInfos(self):
        if self.disablePagination:
            return self.infos.values()
        start = self.currentPage * self.maxPerPage
        end = start + self.maxPerPage
        keys = list(self.infos.keys())[start:end]
        return [self.infos[key] for key in keys]
    
    def getContentView(self):
        return self.contentview
    
    def getInfos(self):
        return self.infos.values()

    def setPaginationPanel(self):
        if self.disablePagination:
            return 
        if self.pagePanel is not None:
            for widget in self.pagePanel.winfo_children():
                widget.grid_forget()
            self.pagePanel.forget()
        self.pagePanel = CTkFrame(self,  height=0) # adjust auto
        btn = ttk.Label(self.pagePanel, text="<<", style="Pagination.TLabel")
        btn.bind('<Button-1>', lambda event:self.goToPage("first"))
        btn.grid(padx=3)
        btn = ttk.Label(self.pagePanel, text="<", style="Pagination.TLabel")
        btn.bind('<Button-1>', lambda event:self.goToPage("previous"))
        btn.grid(row=0, column=1, padx=3)
        col = 2
        start = max(self.currentPage - 2, 0)
        i = start 
        while i <= self.lastPage and i <= start + 5:
            if i == self.currentPage:
                btn = ttk.Label(self.pagePanel, text=str(i), style="CurrentPagination.TLabel")
                btn.grid(column=col,row=0, padx=3)
            else:
                btn = ttk.Label(self.pagePanel, text=str(i), style="Pagination.TLabel")
                btn.bind('<Button-1>', lambda event:self.goToPage(event))
                btn.grid(column=col,row=0, padx=3)
            col +=1
            i += 1
        btn = ttk.Label(self.pagePanel, text=">", style="Pagination.TLabel")
        btn.bind('<Button-1>', lambda event:self.goToPage("next"))
        btn.grid(row=0, column=col, padx=3)
        btn = ttk.Label(self.pagePanel, text=">>", style="Pagination.TLabel")
        btn.bind('<Button-1>', lambda event:self.goToPage("last"))
        btn.grid(row=0, column=col+1, padx=3)
        self.pagePanel.grid(row=1, column=0)

    def resetPagination(self):
        self.currentPage = 0
        self.lastPage = 0
        self.setPaginationPanel()

    def goToPage(self, p, force=False):
        if self.disablePagination:
            return
        if not isinstance(p, int) and not isinstance(p, str):
            p = p.widget.cget("text")
        if p == "first":
            p = 0
        elif p == "last":
            p = self.lastPage
        elif p == "previous":
            p = max(self.currentPage - 1, 0)
        elif p == "next":
            p = min(self.currentPage + 1, self.lastPage)
        else:
            p = int(p)
        if p == self.currentPage and not force:
            return
        self.currentPage = p
        toInsert = self.getShownInfos()
        self.reset_contentview()
        self.insert_items(toInsert)
        self.setPaginationPanel()
        self.update_contentview()

    def reset_contentview(self):
        self.callback_resetview() 


    def insert_items(self, items):
        self.callback_insert(items)

    def filter(self, *args, **kwargs):
        if self._save_infos is not None:
            self.infos = self._save_infos
        subset_items = self._brut_searcher(self.infos, *args, **kwargs)
        self._save_infos = self.infos
        self.infos = OrderedDict()
        self.reset_contentview()
        for item in subset_items:
            self.addPaginatedInfo(self._save_infos[item], auto_update_pagination=False)
        self.setPaginationPanel()
        self.goToPage(0, force=True)
        self.update_contentview()

    def _brut_searcher(self, infos, *args, **kwargs):
        i_r = -1
        ret = []
        check_case = kwargs.get("check_case", True)
        for info_iid, info_value in infos.items():
            allValid = True
            oneValid = False
            for iarg, arg in enumerate(args):
                text = self.callback_get_value(info_value, iarg)
                check_all = kwargs.get("check_all", True)
                if isinstance(arg, str):
                    if not check_case:
                        text = str(text).lower()
                        arg = str(arg).lower()
                    is_valid = arg in str(text)
                
                elif isinstance(arg, bool):
                    is_valid = arg
                elif isinstance(arg, list):
                    is_valid = str(text) in arg
                if not is_valid:
                    allValid = False
                    if check_all:
                        break
                else:
                    oneValid = True
                    if not check_all:
                        break
            if allValid and check_all:
                i_r += 1
                ret.append(info_iid)
            elif oneValid and not check_all:
                i_r += 1
                ret.append(info_iid)
            
        return ret
        


"""Deprecated.
Ask the user to enter a command and select a worker and plugin to launch it."""
import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.application.pollenisatorentry import PopoEntry

class ChildDialogCustomCommand:
    """
    Open a child dialog of a tkinter application to ask details about
    a custom command to launch on target.
    """

    def __init__(self, parent, workers, default_worker="localhost"):
        """
        Open a child dialog of a tkinter application to ask details about
        a custom command to launch on target.

        Args:
            parent: the tkinter parent view to use for this window construction.
            workers: A list of workers registered.
            default_worker: a worker to be selected by default.
        """
        self.app = CTkToplevel(parent)
        self.app.attributes("-type", "dialog")
        self.app.title("Custom command")
        appFrame = CTkFrame(self.app)
        self.app.bind("<Escape>", self.onError)
        self.app.resizable(False, False)
        self.rvalue = None
        self.parent = parent
        lbl = CTkLabel(appFrame, text="Enter the custom command Name")
        lbl.pack()
        self.ent_customCommandName = PopoEntry(appFrame, width=50)
        self.ent_customCommandName.pack()
        lbl = CTkLabel(
            appFrame, text="Enter the custom command to launch")
        lbl.pack()
        self.ent_customCommand = PopoEntry(appFrame, width=50)
        self.ent_customCommand.pack()
        lbl2 = CTkLabel(appFrame, text="Select the parser")
        lbl2.pack()
        apiclient = APIClient.getInstance()
        parsers = apiclient.getPlugins()
        self.box_template = CTkComboBox(
            appFrame, values=tuple([x["plugin"] for x in parsers]), state="readonly")
        self.box_template.set("Default")
        self.box_template.pack()
        lbl3 = CTkLabel(appFrame, text="Select the worker")
        lbl3.pack()
        self.box_workers = CTkComboBox(
            appFrame, values=tuple(workers), state="readonly")
        self.box_workers.set(default_worker)
        self.box_workers.pack()
        self.ok_button = CTkButton(appFrame, text="OK", command=self.onOk)
        self.ok_button.pack(side=tk.BOTTOM, pady=5)
        appFrame.pack(ipady=10, ipadx=10)
        try:
            self.app.wait_visibility()
            self.app.transient(parent)
            self.app.focus_force()
            self.app.grab_set()
            self.app.lift()
        except tk.TclError:
            pass

    def onError(self, event=None):
        self.app.destroy()
        return
    
    def onOk(self):
        """
        Called when the user clicked the validation button. Set the rvalue attributes to the value selected and close the window.
        """
        # send the data to the parent
        self.rvalue = (self.ent_customCommandName.get(), self.ent_customCommand.get(
        ), self.box_template.get(), self.box_workers.get())
        self.app.destroy()



"""
Extend the CTkScrollableFrame to bind also for linux
<MouseWheel> event is not handled by some distributions that only handle <Button-4> and <Button-5> events.
"""

from customtkinter import *
import tkinter as tk 

class ScrollableFrameXPlateform(CTkScrollableFrame):
    def __init__(self, parent, orientation=tk.VERTICAL, **kwargs):
        super().__init__(parent, orientation=orientation, **kwargs)
        self.parent = parent
        self.activate()


    def activate(self):
        self.bind('<Enter>', self.boundToMousewheel)
        self.bind('<Leave>', self.unboundToMousewheel)

    def _onMousewheel(self, event):
        """Scroll the settings canvas
        Args:
            event: scroll info filled when scroll event is triggered"""
        if event.num == 5 or event.delta == -120:
            count = -1
        if event.num == 4 or event.delta == 120:
            count = 1
        try:
            self.event_generate("<MouseWheel>", delta=count)
        except tk.TclError:
            pass
        

    def boundToMousewheel(self, _event=None):
        """Called when the main view canvas is focused.
        Bind the command scrollbar button on linux to the main view canvas
        Args:
            _event: not used but mandatory"""
        self.bind_all("<Button-5>", self._onMousewheel)
        self.bind_all("<Button-4>", self._onMousewheel)

    def unboundToMousewheel(self, _event=None):
        """Called when the main view canvas is unfocused.
        Unbind the command scrollbar button on linux to the main view canvas
        Args:
            _event: not used but mandatory"""
        self.unbind_all("<Button-4>")
        self.unbind_all("<Button-5>")


"""View for interval object. Handle node in treeview and present forms to user when interacted with."""

from pollenisatorgui.core.views.viewelement import ViewElement
from tkinter import TclError


class IntervalView(ViewElement):
    """View for interval object. Handle node in treeview and present forms to user when interacted with.
    Attributes:
        icon: icon name to show in treeview. Icon filename must be in icon directory.
    """

    icon = 'date.png'
    def openModifyWindow(self, **kwargs):
        """
        Creates a tkinter form using Forms classes. This form aims to update or delete an existing Interval
        """
        self.form.clear()
        modelData = self.controller.getData()
        top_panel = self.form.addFormPanel(grid=True)
        top_panel.addFormLabel("Start date")
        top_panel.addFormDate(
            "Start date", self.mainApp, modelData["dated"], column=1)
        top_panel.addFormLabel("End date", row=1)
        top_panel.addFormDate("End date", self.mainApp, modelData["datef"], row=1, column=1)
        top_panel.addFormHelper(
            "Auto scan will only launch if current time fits this interval", row=1, column=2)
        # added in getEmptyModel
        self.form.addFormHidden("waveName", modelData["wave"])
        self.completeModifyWindow()

    def openInsertWindow(self):
        """
        Creates a tkinter form using Forms classes. This form aims to insert a new Interval
        """
        modelData = self.controller.getData()
        top_panel = self.form.addFormPanel(grid=True)
        top_panel.addFormLabel("Start date")
        top_panel.addFormDate("Start date", self.mainApp, "01/01/2000 00:00:00", column=1)
        top_panel.addFormLabel("End date", row=1)
        top_panel.addFormDate(
            "End date", self.mainApp, "31/12/2099 00:00:00", row=1, column=1)
        top_panel.addFormHelper(
            "Auto scan will only launch if current time fits this interval", row=1, column=2)
        # added in getEmptyModel
        self.form.addFormHidden("waveName", modelData["wave"])

        self.completeInsertWindow()

    def getAdditionalContextualCommands(self):
        return {"Insert Interval":self.openInsertWindow}

    def addInTreeview(self, parentNode=None, **kwargs):
        """Add this view in treeview. Also stores infos in application treeview.
        Args:
            parentNode: if None, will calculate the parent. If setted, forces the node to be inserted inside given parentNode.
            _addChildren: not used here
        """
        self.appliTw.views[str(self.controller.getDbId())] = {"view": self}

        parentDbId = parentNode

        if parentNode is None:
            parentNode = self.getParentNode()
        elif "intervals" not in parentNode:
            parentNode = IntervalView.DbToTreeviewListId(parentDbId)
        try:
            parentNode = self.appliTw.insert(
                self.controller.getParentId(), 0, parentNode, text="Intervals", image=self.getClassIcon())
        except TclError:  #  trigger if tools list node already exist
            pass
        try:
            self.appliTw.insert(parentNode, "end", str(
                self.controller.getDbId()), text=str(self.controller.getModelRepr()), tags=self.controller.getTags(), image=self.getClassIcon())
        except:
            pass
        if "hidden" in self.controller.getTags():
            self.hide("tags")
        if self.mainApp.settings.is_checklist_view():
            self.hide("checklist_view")

    def updateReceived(self, obj=None, old_obj=None):
        """Called when a interval update is received by notification.
        Update the interval node and tells to the parent wave to update (its tools).
        """
        if self.controller.model is None:
            return
        parentId = self.controller.getParentId()
        parentView = self.appliTw.views[str(parentId)]["view"]
        parentView.updateReceived()
        super().updateReceived()

    def insertReceived(self):
        """Called when a interval insertion is received by notification.
        Insert the node in treeview.
        Also tells to the parent wave to update (its tools)
        """
        if self.controller.model is None:
            return
        parentId = self.controller.getParentId()
        parentView = self.appliTw.views[str(parentId)]["view"]
        parentView.updateReceived()
        super().insertReceived()

    @classmethod
    def treeviewListIdToDb(cls, treeview_id):
        """Extract from the unique string identifying a list of intervals the parent db ID
        Args:
            treeviewId: the treeview node id of a list of intervals node
        Returns:
            the parent object mongo id as string
        """
        return str(treeview_id).split("|")[1]

    @classmethod
    def DbToTreeviewListId(cls, parent_db_id):
        """Converts a mongo Id to a unique string identifying a list of intervals given its parent
        Args:
            parent_db_id: the parent node mongo ID
        Returns:
            A string that should be unique to describe the parent list of interval node
        """
        return "intervals|"+str(parent_db_id)



"""View for computer object. Handle node in treeview and present forms to user when interacted with."""

from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.views.viewelement import ViewElement
from tkinter import TclError
import tkinter as tk

class ComputerView(ViewElement):
    """View for computer object. Handle node in treeview and present forms to user when interacted with.
    Attributes:
        icon: icon name to show in treeview. Icon filename must be in icon directory.
    """

    icon = 'computer.png'
    def openModifyWindow(self, **kwargs):
        """
        Creates a tkinter form using Forms classes. This form aims to update or delete an existing Computer
        """
        self.form.clear()
        computer_data = self.controller.getData()
        panel = self.form
        panel_info = panel.addFormPanel(grid=True)
        panel_info.addFormLabel("IP")
        panel_info.addFormStr("IP", "", computer_data.get("ip", ""), status="readonly", row=0, column=1)
        panel_info.addFormLabel("Name", column=2)
        panel_info.addFormStr("Name", "", computer_data.get("name", ""), status="readonly", column=3)
        row = 1
        for info_name, val in computer_data.get("infos", {}).items():
            panel_info.addFormLabel(info_name, row=row)
            panel_info.addFormStr(info_name, "", val, status="readonly", row=row, column=1)
            row += 1
        apiclient = APIClient.getInstance()
        res = apiclient.getComputerUsers(str(computer_data["_id"]))
        if res is  None:
            res = {}
        users_data = [(d["domain"], d["username"], d["password"]) for d in res.get("users", [])]
        admins_data = [(d["domain"], d["username"], d["password"]) for d in res.get("admins",[])]
        panel.addFormLabel("Users", side="top")
        panel.addFormTreevw("Users", ("Domain", "Username", "Password"), users_data, fill=tk.X, height=2, max_height=10, side="top")
        
        panel.addFormLabel("Admins", side="top")
        panel.addFormTreevw("Admins", ("Domain", "Username", "Password"), admins_data,  fill=tk.X, height=2, max_height=10, side="top")
        if computer_data.get("secrets", []):
            panel.addFormLabel("Secrets", side="top")
            panel.addFormTreevw("Secrets", ("Secret", ""), [(s, "") for s in computer_data.get("infos", {}).get("secrets", [])], side="top")
        ntds = computer_data.get("ntds", [])
        if ntds:
            panel.addFormLabel("NTDS", side="top")
            panel.addFormText("NTDS", "", "\n".join(ntds), side="top")
        self.completeModifyWindow(editable=False)
        
    def openInsertWindow(self):
        """
        Creates a tkinter form using Forms classes. This form aims to insert a new Computer
        """
        # modelData = self.controller.getData()
        # top_panel = self.form.addFormPanel(grid=True)
        # top_panel.addFormLabel("Start date")
        # top_panel.addFormDate("Start date", self.mainApp, "01/01/2000 00:00:00", column=1)
        # top_panel.addFormLabel("End date", row=1)
        # top_panel.addFormDate(
        #     "End date", self.mainApp, "31/12/2099 00:00:00", row=1, column=1)
        # top_panel.addFormHelper(
        #     "Auto scan will only launch if current time fits this computer", row=1, column=2)
        # # added in getEmptyModel
        # self.form.addFormHidden("waveName", modelData["wave"])

        # self.completeInsertWindow()
        raise NotImplementedError("ComputerView.openInsertWindow not implemented")

    def getAdditionalContextualCommands(self):
        return {}

    def addInTreeview(self, parentNode=None, **kwargs):
        """Add this view in treeview. Also stores infos in application treeview.
        Args:
            parentNode: if None, will calculate the parent. If setted, forces the node to be inserted inside given parentNode.
            _addChildren: not used here
        """

        self.appliTw.views[str(self.controller.getDbId())] = {"view": self}
        tags = self.controller.getTags()
        if parentNode is None:
            parentNode = self.getParentNode()
        try:
            self.appliTw.insert(parentNode, "end", str(
                self.controller.getDbId()), text=str(self.controller.getModelRepr()), tags=tags, image=self.getClassIcon())
        except:
            pass
        if "hidden" in tags:
            self.hide("tags")
        if self.mainApp.settings.is_checklist_view():
            self.hide("checklist_view")

   


"""Describe tkinter Label as image with default common args"""

import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
from pollenisatorgui.core.forms.form import Form
from PIL import ImageTk, Image


class FormImage(Form):
    """
    Form field representing a label Image.
    Default setted values:
        if pack : padx = pady = 5, side = left
        if grid: row = column = 0 sticky = "East"
    """
    def __init__(self, path, **kwargs):
        """
        Constructor for a form label
        Args:
            path: the path to the image
            kwargs: same keyword args as you would give to CTkLabel
        """
        super().__init__("Image")
        self.path = path
        self.kwargs = kwargs

    def constructView(self, parent):
        """
        Create the label view inside the parent view given

        Args:
            parent: parent FormPanel.
        """
        self.img = Image.open(self.path)
        self.tkimage = CTkImage(self.img)
        self.lbl = CTkLabel(parent.panel, text="", image=self.tkimage, justify=tk.LEFT)
        if parent.gridLayout:
            self.lbl.grid(column=self.getKw("column", 0), row=self.getKw("row", 0), sticky=self.getKw("sticky", tk.E) , padx=self.getKw("padx", 5), pady=self.getKw("pady", 5), **self.kwargs)
        else:
            self.lbl.pack(side=self.getKw("side", "left"), padx=self.getKw("padx", 10), pady=self.getKw("pady", 5), **self.kwargs)

    
    def setImage(self, newPath):
        """nothing to set so overwrite
        Args:
            _newval: not used"""
        self.path = newPath
        self.img = Image.open(newPath)
        self.tkimage = CTkImage(self.img)
        self.lbl.configure(image=self.tkimage)
        self.lbl.image = self.tkimage



"""Controller for defect object. Mostly handles conversion between mongo data and python objects"""

import os
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.controllers.controllerelement import ControllerElement
import pollenisatorgui.core.models.defect as defect
import re

class DefectController(ControllerElement):
    """Inherits ControllerElement
    Controller for defect object. Mostly handles conversion between mongo data and python objects"""

    def doUpdate(self, values):
        """
        Update the Defect represented by this model in database with the given values.

        Args:
            values: A dictionary crafted by DefectView containg all form fields values needed.

        Returns:
            The mongo ObjectId _id of the updated Defect document.
        """
        self.model.title = values.get("Title", self.model.title)
        self.model.synthesis = values.get("Synthesis", self.model.synthesis)
        self.model.description = values.get("Description", self.model.description)
        self.model.ease = values.get("Ease", self.model.ease)
        self.model.impact = values.get("Impact", self.model.impact)
        self.model.risk = values.get("Risk", self.model.risk)
        self.model.redactor = values.get("Redactor", self.model.redactor)
        mtype = values.get("Type", None)
        if mtype is not None:
            mtype = [k for k, v in mtype.items() if v == 1]
            self.model.mtype = mtype
        self.model.language = values.get("Language", self.model.language)
        self.model.notes = values.get("Notes", self.model.notes)
        fixes_inserted = values.get("Fixes", None)
        if fixes_inserted is not None:
            self.model.fixes = []
            for fix in fixes_inserted:
                if fix[0].strip() == "":
                    continue
                self.model.fixes.append({"title": fix[0], "execution": fix[1], "gain": fix[2],
                                          "synthesis": fix[3], "description": fix[4]})
        self.model.infos = values.get("Infos", self.model.infos)
        for info in self.model.infos:
            self.model.infos[info] = self.model.infos[info][0]

        group_proofs = self.findProofsInDescription(self.model.description)
        for group_proof in group_proofs:
            matched = group_proof.group(0)
            path = group_proof.group(1)
            if path.strip() != "":
                result_path = self.model.uploadProof(path)
                if result_path is not None:
                    result_path = os.path.basename(result_path)
                    self.model.description = self.model.description.replace(matched, f"![{result_path}]({result_path})")
        # Updating
        self.model.update()

    def update_fixes(self,fixes):
        """Update the fixes of the model"""
        self.model.fixes = fixes
        self.model.update({"fixes": fixes})

    def findProofsInDescription(self, description):
        regex_images = r"!\[.*\]\(((?!http).*)\)"
        return re.finditer(regex_images, description)
        

    def doInsert(self, values):
        """
        Insert the Defect represented by this model in the database with the given values.

        Args:
            values: A dictionary crafted by DefectView containing all form fields values needed.

        Returns:
            {
                '_id': The mongo ObjectId _id of the inserted command document.
                'nbErrors': The number of objects that has not been inserted in database due to errors.
            }
        """
        title = values["Title"]
        synthesis = values.get("Synthesis", "")
        description = values.get("Description", "")
        ease = values["Ease"]
        impact = values["Impact"]
        redactor = values["Redactor"]
        mtype_dict = values["Type"]
        language = values["Language"]
        risk = values["Risk"]
        notes = values.get("Notes", "")
        mtype = [k for k, v in mtype_dict.items() if v == 1]
        target_id = values.get("target_id", "")
        target_type = values.get("target_type", "")
        fixes = []
        fixes_inserted = values.get("Fixes", None)
        if fixes_inserted is not None:
            for fix in fixes_inserted:
                if fix[0].strip() == "":
                    continue
                fixes.append({"title": fix[0], "execution": fix[1], "gain": fix[2],
                                          "synthesis": fix[3], "description": fix[4]})
        tableau_from_ease = defect.Defect.getTableRiskFromEase()
       
        if risk == "" or risk == "N/A":
            risk = tableau_from_ease.get(ease,{}).get(impact,"N/A")
        self.model.initialize(target_id, target_type, title, synthesis, description, ease,
                              impact, risk, redactor, mtype, language, notes, None, fixes, [])
        ret, _ = self.model.addInDb()
        # Update this instance.
        # Upload proof after insert on db cause we need its mongoid
        group_proofs = self.findProofsInDescription(description)
        must_update = False
        apiclient = APIClient.getInstance()
        for group_proof in group_proofs:
            matched = group_proof.group(0)
            path = group_proof.group(1)
            if path.strip() != "" and not path.startswith("http"):
                result_path = self.model.uploadProof(path)
                if result_path is not None:
                    result_path = os.path.basename(result_path)
                    self.model.description = self.model.description.replace(matched, f"![{result_path}]({result_path})")
                    must_update = True
        
        if must_update:
            self.model.update({"description": self.model.description}) 

        return ret, 0  # 0 erros

    def addAProof(self, proof_path):
        """Add a proof file to model defect.
        Args:
            formValues: the view form values as a dict. Key "Proof "+str(index) must exist
        """
        if proof_path.strip() == "":
            return
        result_path = self.model.uploadProof(proof_path)
        self.model.update()
        return result_path

    def getProof(self, ind):
        """Returns proof file to model defect.
        Args:
            ind: the proof index in the form to get
        Returns:
            the local path of the downloaded proof (string)
        """
        return self.model.getProof(ind)
    
    def getProofWithName(self, remote_name):
        """Returns proof file to model defect.
        Args:
            remote_name: the proof remote name to get
        Returns:
            the local path of the downloaded proof (string)
        """
        return os.path.realpath(self.model.getProofWithName(remote_name))

    def deleteProof(self, ind):
        """Delete a proof file given a proof index
        Args:
            ind: the proof index in the form to delete
        """
        self.model.removeProof(ind)

    def isAssigned(self):
        """Checks if the defect model is assigned to an IP or is global
        Returns:    
            bool
        """
        return self.model.isAssigned()
    
    def getTargetRepr(self):
        """Returns the target representation of the defect model
        Returns:
            str
        """
        return self.model.getTargetRepr()

    
    def getType(self):
        """Returns a string describing the type of object
        Returns:
            "defect" """
        return "defect"


"""Widget with no display that holds a value"""
from pollenisatorgui.core.forms.form import Form


class FormHidden(Form):
    """
    Form field hidden, to store a value.
    """

    def __init__(self, name, default=""):
        """
        Constructor for a hidden form.

        Args:
            name: the form name.
            default: a default value to store in it.
        """
        super().__init__(name)
        self.default = default
        self.val = default

    def getValue(self):
        """
        Return the form value. Required for a form.

        Returns:
            Return the form value.
        """
        return self.val

    def setValue(self, newval):
        """
        Set the form value. Required for a form.
        Args:
           newval: new value to be setted
        """
        self.val = newval



"""This class pop a remark view form in a subdialog"""

import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
from pollenisatorgui.core.views.remarkview import RemarkView
from pollenisatorgui.core.controllers.remarkcontroller import RemarkController
from pollenisatorgui.core.models.remark import Remark
import pollenisatorgui.core.components.utils as utils

class ChildDialogRemarkView:
    """
    Open a child dialog of a tkinter application to answer a question.
    """
    def __init__(self, parent, remarkModel=None):
        """

        Args:
            parent: the tkinter parent view to use for this window construction.
            remarkModel : A Remark Model object to load default values. None to have empty fields, default is None.
        """
        self.app = CTkToplevel(parent)
        self.app.attributes("-type", "dialog")
        self.isInsert = remarkModel is None
        if self.isInsert:
            remarkModel = Remark()
        title = "Add a remark" if self.isInsert else "Edit remark"
        self.app.title(title)
        self.app.resizable(False, False)
        self.rvalue = None
        appFrame = CTkFrame(self.app)
        self.app.bind("<Escape>", self.cancel)
        
        self.remark_vw = RemarkView(appFrame, RemarkController(remarkModel))
        if self.isInsert:
            self.remark_vw.openInsertWindow(addButtons=False)
        else:
            self.remark_vw.openModifyWindow(addButtons=False)
        ok_button = CTkButton(appFrame, text="OK")
        ok_button.pack(side="right", padx=5, pady=10)
        ok_button.bind('<Button-1>', self.okCallback)
        cancel_button = CTkButton(appFrame, text="Cancel", 
                               fg_color=utils.getBackgroundColor(), text_color=utils.getTextColor(),
                               border_width=1, border_color="firebrick1", hover_color="tomato")
        cancel_button.pack(side="right", padx=5, pady=10)
        cancel_button.bind('<Button-1>', self.cancel)
        appFrame.pack(fill=tk.BOTH, ipady=10, ipadx=10)
        self.app.transient(parent)
        try:
            self.app.wait_visibility()
            self.app.transient(parent)
            self.app.grab_set()
            self.app.focus_force()
            self.app.lift()
        except tk.TclError:
            pass

    def cancel(self, _event=None):
        """called when canceling the window.
        Close the window and set rvalue to False
        Args:
            _event: Not used but mandatory"""
        self.rvalue = False
        self.app.destroy()

    def okCallback(self, _event=None):
        """called when pressing the validating button
        Close the window if the form is valid.
        Set rvalue to True and perform the remark update/insert if validated.
        Args:
            _event: Not used but mandatory"""
        
        if self.isInsert:
            res, _ = self.remark_vw.insert()
        else:
            res, _ = self.remark_vw.update()
        if res:
            self.rvalue = True
            self.app.destroy()



"""Ttk treeview class with added functions.
"""
import tkinter as tk
from bson.objectid import ObjectId
from pollenisatorgui.core.components.datamanager import DataManager
from pollenisatorgui.core.models.command import Command
from pollenisatorgui.core.views.commandview import CommandView
from pollenisatorgui.core.views.multicommandview import MultiCommandView
from pollenisatorgui.core.controllers.commandcontroller import CommandController
from pollenisatorgui.core.application.treeviews.PollenisatorTreeview import PollenisatorTreeview
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.application.dialogs.ChildDialogQuestion import ChildDialogQuestion
import pollenisatorgui.core.components.utils as utils

class CommandsTreeview(PollenisatorTreeview):
    """CommandsTreeview class
    Inherit PollenisatorTreeview.
    Ttk treeview class with added functions to handle the command objects.
    """

    def __init__(self, appli, parentFrame):
        """
        Args:
            appli: a reference to the main Application object.
            parentFrame: the parent tkinter window object.
        """
        super().__init__(appli, parentFrame)
        self.commands_node = None  # parent of all commands nodes
        self.openedViewFrameId = None  # if of the currently opened object in the view frame
        
    

    def initUI(self, _event=None):
        """Initialize the user interface widgets and binds them.
        Args:
            _event: not used but mandatory
        """
        if self.commands_node is not None:
            return
        self._initContextualsMenus()
        self.heading('#0', text='Commands', anchor=tk.W)
        self.column('#0', stretch=tk.YES, minwidth=300, width=300)
        self.bind("<Button-3>", self.doPopup)
        self.bind("<<TreeviewSelect>>", self.onTreeviewSelect)
        #self.bind("<Return>", self.onTreeviewSelect)
        #self.bind("<Button-1>", self.onTreeviewSelect)
        self.bind('<Delete>', self.deleteSelected)
        self.load()
        self.switchExpandCollapse(True, self.commands_node)

    def onTreeviewSelect(self, event=None):
        """Called when a line is selected on the treeview
        Open the selected object view on the view frame.
        IF it's a parent commands or other node, opens Insert
        ELSE open a modify window
        Args:
            event: filled with the callback, contains data about line clicked
        """
        selection = self.selection()
        if len(selection) == 1:
            item = super().onTreeviewSelect(event)
            if isinstance(item, str):
                pass
                # apiclient = APIClient.getInstance()
                # if str(item) == "mycommands":
                #     user = apiclient.getUser()
                #     objView = CommandView(
                #         self, self.appli.commandsViewFrame, self.appli, CommandController(Command({"owners":[user]})))
                #     objView.openInsertWindow()
                # elif str(item) == "commands":
                #     objView = CommandView(
                #         self, self.appli.commandsViewFrame, self.appli, CommandController(Command({})))
                #     objView.openInsertWindow()
            else:
                self.openModifyWindowOf(item)
        elif len(selection) > 1:
            # Multi select:
            multiView = MultiCommandView(self, self.appli.commandsViewFrame, self.appli)
            for widget in self.appli.commandsViewFrame.winfo_children():
                widget.destroy()
            multiView.form.clear()
            multiView.openModifyWindow()


    def doPopup(self, event):
        """Open the popup 
        Args:
            event: filled with the callback, contains data about line clicked
        """
        self.contextualMenu.selection = self.identify(
            "item", event.x, event.y)
        super().doPopup(event)

    def openModifyWindowOf(self, dbId):
        """
        Retrieve the View of the database id given and open the modifying form for its model and open it.

        Args:
            dbId: the database Mongo Id to modify.
        """
        objView = self.getViewFromId(str(dbId))
        if objView is not None:
            for widget in self.appli.commandsViewFrame.winfo_children():
                widget.destroy()
            objView.form.clear()
            self.openedViewFrameId = str(dbId)
            objView.openModifyWindow(treevw="commands")

    def load(self, _searchModel=None):
        """
        Load the treeview with database information

        Args:
            _searchModel: (Deprecated) inherited not used. 
        """
        for widget in self.appli.commandsViewFrame.winfo_children():
            widget.destroy()
        self.delete(*self.get_children())

        self._load()

    def _load(self):
        """
        Load the treeview with database information
        """
        self.commands_node = self.insert(
            "", "end", "commands", text="Commands", image=CommandView.getClassIcon())
        commands = Command.fetchObjects({})
        for command in commands:
            command_vw = CommandView(
                self, self.appli.commandsViewFrame, self.appli, CommandController(command))
            command_vw.addInTreeview()
        
    def deleteSelected(self, _event=None):
        """
        Interface to delete a database object from an event.
        Prompt the user a confirmation window.
        Args:
            _event: not used, a ttk Treeview event autofilled. Contains information on what treeview node was clicked.
        """
        n = len(self.selection())
        dialog = ChildDialogQuestion(self.parentFrame,
                                     "DELETE WARNING", "Becareful for you are about to delete "+str(n) + " entries and there is no turning back.", ["Delete", "Cancel"])
        self.wait_window(dialog.app)
        if dialog.rvalue != "Delete":
            return
        if n == 1:
            view = self.getViewFromId(self.selection()[0])
            if view is None:
                return
            view.delete(None, False)
        else:
            toDelete = {}
            for selected in self.selection():
                view = self.getViewFromId(selected)
                if view is not None:
                    viewtype = view.controller.model.coll_name
                    if viewtype not in toDelete:
                        toDelete[viewtype] = []
                    toDelete[viewtype].append(view.controller.getDbId())
            apiclient = APIClient.getInstance()
            apiclient.bulkDeleteCommands(toDelete)

    def refresh(self, force=True):
        """Alias to self.load method"""
        self.load()

    def _initContextualsMenus(self):
        """
        Create the contextual menu
        """
        
        self.contextualMenu = utils.craftMenuWithStyle(self.parentFrame)
        self.contextualMenu.add_command(
            label="Add command", command=self.addCommand)
        self.contextualMenu.add_command(
            label="Duplicate command", command=self.duplicateCommand)
        self.contextualMenu.add_separator()
        self.contextualMenu.add_command(
            label="Sort children", command=self.sort)
        self.contextualMenu.add_command(
            label="Expand", command=self.expand)
        self.contextualMenu.add_command(
            label="Collapse", command=self.collapse)
        self.contextualMenu.add_separator()
        self.contextualMenu.add_command(
            label="Close", command=self.closeMenu)
        super()._initContextualsMenus
        return self.contextualMenu
    
    def addCommand(self, _event=None):
        objView = CommandView(
                        self, self.appli.commandsViewFrame, self.appli, CommandController(Command({})))
        objView.openInsertWindow()

    def duplicateCommand(self, _event=None):
        for selected in self.selection():
            view_o = self.getViewFromId(selected)
            if view_o is not None:
                if isinstance(view_o, CommandView):
                    data = view_o.controller.getData()
                    del data["_id"]
                    data["name"] = data["name"] + "_dup"
                    objView = CommandView(
                        self, self.appli.commandsViewFrame, self.appli, CommandController(Command(data)))
                    objView.openInsertWindow()
                    
    
    def update_received(self, dataManager, notification, obj, oldObj):
        """
        Callback for the observer pattern implemented in mongo.py.

        Args:
            collection: the collection that has been modified
            iid: the mongo ObjectId _id that was modified/inserted/deleted
            action: update/insert/delete. It was the action performed on the iid
            parent: the mongo ObjectId of the parent. Only if action in an insert.
        """
        if notification["db"] != "pollenisator":
            return
        if notification["collection"] != "commands":
            return
        action = notification["action"]
        iid = notification["iid"]
        # Delete
        apiclient = APIClient.getInstance()
        if action == "delete":
            try:
                self.delete(ObjectId(iid))
            except tk.TclError:
                pass  # item was not inserted in the treeview
        # Insert
        if action == "insert":
            res = Command.fetchObject({"_id": ObjectId(iid)})
            view = CommandView(self, self.appli.commandsViewFrame,
                                self.appli, CommandController(res))
            parent = None
            try:
                view.addInTreeview(parent)
                if view is not None:
                    view.insertReceived()
            except tk.TclError:
                pass
        if action == "update":
            try:
                view = self.getViewFromId(str(iid))
                if view is not None:
                    self.item(str(iid), text=str(
                        view.controller.getModelRepr()), image=view.getIcon())
            except tk.TclError:
                if view is not None:
                    view.addInTreeview()
            if str(iid) == self.openedViewFrameId:
                self.event_generate("<F6>")
            if view is not None:
                view.controller.actualize()
                view.updateReceived()



"""Tag Model. Tag other stuff with it"""

from pollenisatorgui.core.components.datamanager import DataManager
from pollenisatorgui.core.models.element import Element
from pollenisatorgui.core.components.apiclient import APIClient
import pollenisatorgui.core.components.utils as utils
from bson.objectid import ObjectId


class Tag(Element):
    """
    Represents an interval object that defines an time interval where a wave can be executed.

    Attributes:
        coll_name: collection name in pollenisator database
    """
    coll_name = "tags"

    def __init__(self, valuesFromDb=None):
        """Constructor
        Args:
            valueFromDb: a dict holding values to load into the object. A mongo fetched interval is optimal.
                        possible keys with default values are : _id (None), item_id, item_type, tags([]), date
        """
        if valuesFromDb is None:
            valuesFromDb = {}
        super().__init__(valuesFromDb.get("_id", None), None, {})
        self.initialize(valuesFromDb.get("item_id", None), valuesFromDb.get("item_type", None), valuesFromDb.get("tags",[]), valuesFromDb.get("date", None))

    def initialize(self, item_id, item_type, tags, date):
        """Set values of interval
        Args:
            wave: the parent wave name
            dated: a starting date and tiem for this interval in format : '%d/%m/%Y %H:%M:%S'. or the string "None"
            datef: an ending date and tiem for this interval in format : '%d/%m/%Y %H:%M:%S'. or the string "None"
            infos: a dictionnary with key values as additional information. Default to None
        Returns:
            this object
        """
        self.item_id = item_id
        self.item_type = item_type
        self.tags = tags
        self.date = date
        return self

    def delete(self):
        """
        Delete the Interval represented by this model in database.
        """
        apiclient = APIClient.getInstance()
        apiclient.delete(
            "tags", self._id)
        
    @staticmethod
    def addTagTo(item_id, item_type, tag, overrideGroups=True):
        apiclient = APIClient.getInstance()
        apiclient.addTag(item_id, item_type, tag, overrideGroups)
    
    @staticmethod
    def delTag(item_id, item_type, tag):
        apiclient = APIClient.getInstance()
        apiclient.delTag(item_id, item_type, tag)

    @staticmethod
    def setTags(item_id, item_type, tags):
        apiclient = APIClient.getInstance()
        apiclient.setTags(item_id, item_type, tags)

    def getData(self):
        """Return interval attributes as a dictionnary matching Mongo stored intervals
        Returns:
            dict with keys _id, item_id, item_type, tags, date
        """
        return {"item_type": self.item_type, "item_id": self.item_id, "date": self.date, "_id": self.getId(), "tags": self.tags}





from customtkinter import *
from tkinter import ttk
import tkinter as tk
from PIL import Image
import base64
from io import BytesIO


class CollapsibleFrame(CTkFrame):
    def __init__(self, master, text=None, border_width=2, width=0, height=16, interior_padx=0, interior_pady=8, background=None, caption_separation=4, caption_font=None, caption_builder=None, icon_x=5, **kwargs):
        CTkFrame.__init__(self, master, **kwargs)
        
        self._is_opened = False

        self._interior_padx = interior_padx
        self._interior_pady = interior_pady

        self._iconOpen = "R0lGODlhEAAQAKIAAP///9TQyICAgEBAQAAAAAAAAAAAAAAAACwAAAAAEAAQAAADNhi63BMgyinFAy0HC3Xj2EJoIEOM32WeaSeeqFK+say+2azUi+5ttx/QJeQIjshkcsBsOp/MBAA7"
        self._iconClose = "R0lGODlhEAAQAKIAAP///9TQyICAgEBAQAAAAAAAAAAAAAAAACwAAAAAEAAQAAADMxi63BMgyinFAy0HC3XjmLeA4ngpRKoSZoeuDLmo38mwtVvKu93rIo5gSCwWB8ikcolMAAA7"
        self._iconOpen = Image.open(BytesIO(base64.b64decode(self._iconOpen)))
        self._iconClose = Image.open(BytesIO(base64.b64decode(self._iconClose)))
        height_of_icon = max(self._iconOpen.height, self._iconClose.height)
        width_of_icon = max(self._iconOpen.width, self._iconClose.width)
        
        containerFrame_pady = (height_of_icon//2) +1

        self._height = height
        self._width = width

        self._containerFrame = CTkFrame(self, border_width=border_width, width=width, height=height)
        self._containerFrame.pack(expand=True, fill=tk.X, pady=(containerFrame_pady,0))
        
        self.interior = CTkFrame(self._containerFrame)

        self._collapseButton = CTkLabel(self,  text= "", image=CTkImage(self._iconOpen))
        self._collapseButton.place(in_= self._containerFrame, x=icon_x, y=-(height_of_icon//2)-5, anchor=tk.NW, bordermode="ignore")
        self._collapseButton.bind("<Button-1>", lambda _event=None: self.toggle())

        if caption_builder is None:
            self._captionLabel = CTkLabel(self, anchor=tk.W, text=text)
            if caption_font is not None:
                self._captionLabel.configure(font=caption_font)
        else:
            self._captionLabel = caption_builder(self)
            
            if not isinstance(self._captionLabel, ttk.Widget):
                raise Exception("'caption_builder' doesn't return a tkinter widget")

        self.after(0, lambda _event=None: self._place_caption(caption_separation, icon_x, width_of_icon))

    def update_width(self, width=None):
        # Update could be devil
        # http://wiki.tcl.tk/1255
        self.after(0, lambda width=width:self._update_width(width))

    def _place_caption(self, caption_separation, icon_x, width_of_icon):
        try:
            self.update()
            x = caption_separation + icon_x + width_of_icon
            y = -(self._captionLabel.winfo_reqheight()//2)-1
            self._captionLabel.place(in_= self._containerFrame, x=x, y=y, anchor=tk.NW, bordermode="ignore")
        except tk.TclError:
            pass
        
    def _update_width(self, width):
        self.update()
        if width is None:
            width=self.interior.winfo_reqwidth()

        if isinstance(self._interior_pady, (list, tuple)):
            width += self._interior_pady[0] + self._interior_pady[1]
        else:
            width += 2*self._interior_pady
            
        width = max(self._width, width)

        self._containerFrame.configure(width=width)
        
    def open(self):
        self._collapseButton.configure(image=CTkImage(self._iconClose))
        
        self._containerFrame.configure(height=self.interior.winfo_reqheight())
        self.interior.pack(expand=True, fill=tk.X, padx=self._interior_padx, pady =self._interior_pady)

        self._is_opened = True

    def close(self):
        self.interior.pack_forget()
        self._containerFrame.configure(height=self._height)
        self._collapseButton.configure(image=CTkImage(self._iconOpen))
        self._is_opened = False
    
    def toggle(self):
        if self._is_opened:
            self.close()
        else:
            self.open()



import subprocess
import pollenisatorgui.core.components.utils as utils
from pollenisatorgui.core.application.dialogs.ChildDialogCombo import ChildDialogCombo
import psutil
from pollenisatorgui.core.components.apiclient import APIClient
import os
import re

def main(apiclient, appli, **kwargs):
    APIClient.setInstance(apiclient)
    addrs = psutil.net_if_addrs()
    print(addrs.keys())
    dialog = ChildDialogCombo(None, addrs.keys(), displayMsg="Choose your ethernet device connected to a domain")
    dialog.app.wait_window(dialog.app)
    nmcli_path = utils.which_expand_alias("nmcli")
    if nmcli_path is None:
        return False, "nmcli is not installed"
    probable_dc = []
    probable_domain = []
    if dialog.rvalue is not None:
        cmd = f"{nmcli_path} dev show {dialog.rvalue}"
        result = subprocess.check_output(cmd, shell=True)
        result = result.decode("utf-8")
        regex_res = re.findall(r"\.DNS\[\d+\]:\s+(\S+)$", result, re.MULTILINE)
        if regex_res:
            probable_dc += regex_res
        regex_res = re.findall(r"\.DOMAIN\[\d+\]:\s+(\S+)$", result, re.MULTILINE)
        if regex_res:
            probable_domain += regex_res
        msg = ""
        if probable_dc:
            msg += "Probable DC founds: "+"\n".join(probable_dc)
        if probable_domain:
            msg += "\nProbable domain found: "+"\n".join(probable_domain)
        if msg == "":
            msg = "No IP found."
        return True, msg
    return None



"""A registry for all subclasses of Plugin"""
REGISTRY = {}

def register_class(target_class):
    """Register the given class
    Args:
        target_class: type <class>
    """
    REGISTRY[target_class.__name__] = target_class()


class MetaPlugin(type):
    def __new__(meta, name, bases, class_dict):
        cls = type.__new__(meta, name, bases, class_dict)
        if name not in REGISTRY:
            register_class(cls)
        return cls


class Plugin(metaclass=MetaPlugin):
    """ Parent base plugin to be inherited
    Attributes:
        autoDetect: indicating to auto-detect that this plugin is able to auto detect.
    """

    def getActions(self, _toolmodel):
        """
        Summary: Add buttons to the tool view.
        Args:
            * toolmodel : the tool model opened in the pollenisator client.
        Return:
            A dictionary with buttons text as key and function callback as value.
        """
        return {}



from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.components.datamanager import DataManager
from pollenisatorgui.core.models.element import Element

class Share(Element):
    coll_name = "shares"

    def __init__(self, valuesFromDb=None):
        if valuesFromDb is None:
            valuesFromDb = {}
        super().__init__(valuesFromDb.get("_id", None), valuesFromDb.get("parent", None),  valuesFromDb.get("infos", {}))
        self.initialize(valuesFromDb.get("_id"),  valuesFromDb.get("ip"),
            valuesFromDb.get("share"),  valuesFromDb.get("files"), valuesFromDb.get("infos"))

    def initialize(self,  _id=None, ip=None, share=None,  files=None, infos=None): 
        """
        :param ip: The ip of this Share. 
        :type ip: str
        :param share: The share of this Share. 
        :type share: str

        :param files: The files of this Share. 
        :type files: List[ShareFile]
        """
       
        self._id = _id
        self.ip = ip
        self.share = share
        self.files = [] 
        self.infos =  infos if infos is not None else {}
        if files is not None:
            for f in files:
                self.files.append(f)
        return self
    
    def __str__(self):
        """
        Get a string representation of a defect.

        Returns:
            Returns the defect +title.
        """
        return f"{self.ip}\\{self.share} ({str(len(self.files))})"

    def getData(self):
        return {"_id": self._id, "ip":self.ip, "share": self.share,  "files":[f for f in self.files], "infos":self.infos}

    def _getParentId(self):
        """
        Return the mongo ObjectId _id of the first parent of this object. For a share it is the ip.

        Returns:
            Returns the parent share ip's ObjectId _id".
        """
        datamanager = DataManager.getInstance()
        obj = datamanager.find("ips", {"ip":self.ip}, False)
        if obj is None:
            return None
        return obj.getId()

    @classmethod
    def fetchObjects(cls, pipeline):
        """Fetch many commands from database and return a Cursor to iterate over model objects
        Args:
            pipeline: a Mongo search pipeline (dict)
        Returns:
            Returns a cursor to iterate on model objects
        """
        apiclient = APIClient.getInstance()
        pipeline["type"] = "share"
        ds = apiclient.find(cls.coll_name, pipeline, True)
        if ds is None:
            return None
        for d in ds:
            # disabling this error as it is an abstract function
            yield cls(d)  #  pylint: disable=no-value-for-parameter
    
    @classmethod
    def fetchObject(cls, pipeline):
        """Fetch many commands from database and return a Cursor to iterate over model objects
        Args:
            pipeline: a Mongo search pipeline (dict)
        Returns:
            Returns a cursor to iterate on model objects
        """
        apiclient = APIClient.getInstance()
        pipeline["type"] = "share"
        d = apiclient.find(cls.coll_name, pipeline, False)
        if d is None:
            return None
        return cls(d)
    
    @classmethod
    def fetchPentestObjects(cls):
        return [x for x in Share.fetchObjects({})]


"""Module using Lark parser to parse filter query and return according results"""
from lark import Lark, Transformer, exceptions
import re
from bson import ObjectId

from pollenisatorgui.core.components.datamanager import DataManager
from pollenisatorgui.core.components.tag import TagInfos

class Term:
    """A search term, meaning "key.name" == (value) """
    def __init__(self, val):
        """Constructor"""
        self.val = val

class TreeToCondition(Transformer):
    """Inherits lark.Transformer
    A Lark Transformer to process the parse-tree returned by lark
    Attributes:
        null: value will be converted to python None
        true: value will be converted to python True
        false: value will be converted to python False
        eq: value will be converted to string "=="
        neq: value will be converted to string "!="
        gt: value will be converted to string ">"
        ge: value will be converted to string ">="
        le: value will be converted to string "<="
        lt: value will be converted to string "<"
        regex: value will be converted to string "||regex||"
        inside: value will be converted to string "in"
        notin: value will be converted to string "not in"
        andcond: value will be converted to string "and"
        orcond: value will be converted to string "or"
        notcond: value will be converted to string "not"
    """
    null = lambda self, _: None
    true = lambda self, _: True
    false = lambda self, _: False
    eq = lambda self, _: "=="
    neq = lambda self, _: "!="
    gt = lambda self, _: ">"
    ge = lambda self, _: ">="
    le = lambda self, _: "<="
    lt = lambda self, _: "<"
    regex = lambda self, _: "||regex||"
    inside = lambda self, _: "in"
    notin = lambda self, _: "not in"
    andcond = lambda self, _: "and"
    orcond = lambda self, _: "or"
    notcond = lambda self, _: "not"

    def term(self, items):
        """Applied on parse-tree terms objects.
        Args:
            items: the parse-tree term object
        Returns:
            the given item as a list
        """
        return list(items)

    def var(self, s):
        """Applied on parse-tree var objects.
        Args:
            s: the parse-tree var object
        Returns:
            the given item as a Term
        """
        (s,) = s
        return Term(s)

    def string(self, s):
        """Applied on parse-tree string objects.
        Args:
            s: the parse-tree string object
        Returns:
            the given item as a str
        """
        (s,) = s
        return str(s)
    def number(self, n):
        """Applied on parse-tree number objects.
        Args:
            s: the parse-tree number object
        Returns:
            the given item as a str with double quotes around them
        """
        (n,) = n
        return "\""+str(n)+"\""



class ParseError(Exception):
    """Inherits Exception
    Class to raise parsing error"""
    pass

class Filter:
    """
    Class to perform Lark parsing and filter database search.
    Attributes:
        exact_match: (DEPRECATED use in, or regex to perform partial match) default to True, can be changed in constructor
        condition_parser: The parsing syntax of Lark is used to search
                        * term: is a (var == value)   
                        * uniopcond: is unary operation (not)
                        * opcond: logical operator on temrs ("and" and "or")       
                        * opregex: the regex operator
                        * STRING: an alphanumeric string with extras characs '.', '[' and ']'
    """
    exact_match = True
    condition_parser = Lark(r"""
    ?term: "("fixedvalue op fixedvalue")"
                | fixedvalue op fixedvalue
                | STRING opregex ESCAPED_STRING
                | uniopcond term
                | "("uniopcond term")"
                | term opcond term
                | "("term opcond term")"
    uniopcond: "not" -> notcond
    opcond: "and" -> andcond | "or" -> orcond
    op: "==" -> eq | "!=" -> neq | ">" -> gt | "<" -> lt 
        | "<=" -> le | ">=" -> ge | "in" -> inside | "not in" -> notin 
    opregex: "regex" -> regex
    STRING: /[A-Za-z0-9\.\[\]]+/
    fixedvalue: SIGNED_NUMBER -> number
         | "true" -> true
         | "false" -> false
         | "null" -> null
         | ESCAPED_STRING -> string
         | STRING -> var

    %import common.ESCAPED_STRING
    %import common.SIGNED_NUMBER
    %import common.WS
    %ignore WS
    """, start='term', parser="lalr", transformer=TreeToCondition())

    @classmethod
    def help(cls):
        """Return a string to help typing request by providing examples
        Returns:
            A string of examples
        """
        # pylint: disable=anomalous-backslash-in-string
        return """
In Text Search mode, results are text based only.

In Advanced mode (not text search), filters are used

Search examples in match (python type condition):

type == "port"
type == "port" and port == 443
type == "port" and port regex "443$"
type == "port" and (port == 80 or port == 443)
type == "port" and port != 443
type == "port" and port != 443 and port != 80
type == "defect"
type == "defect" and "Foo" in title
type == "ip" and ip regex "[A-Za-z]"
type == "ip" and ip regex "^1\.2"
type == "tool" and "done" in status
type == "tool" and "done" not in status
type == "tool" and "ready" in status
type == "ip" and infos.key == "ABC"
port == 443
"""
    def __init__(self, query="", exactMatch=True):
        """Constructor
        Args:
            query: the query to parse
            exactMatch: not used
        Raises:
            ParseError if Lark raises an UnexpectedToken or an UnexptectedCharacters exception.
        """
        Filter.exact_match = exactMatch
        try:
            self.parsed = Filter.condition_parser.parse(query)
        except exceptions.UnexpectedToken as e:
            raise ParseError(e)
        except exceptions.UnexpectedCharacters as e:
            raise ParseError(e)

    def evaluate(self, parsedcopy, data):
        """Replace parsed items with corresponding formated data and eval the condition using python eval()
        Args:
            parsedcopy: parsed items generated by Lark
            data: data of any object as dictionnary.
        Returns:
            returns result of final eval function
        """
        phrase = ""
        if "||regex||" in parsedcopy:
            val = data.get(parsedcopy[0], None)
            if val is None:
                return False
            return re.search(str(parsedcopy[2])[1:-1], str(val)) is not None
        for children in parsedcopy:
            if isinstance(children, Term):
                if children.val == "id":
                    children.val = "_id"
                if children.val in data.keys():
                    val = data[children.val]
                    if isinstance(val, str):
                        val = "\"\"\""+val.lower()+"\"\"\""
                    elif isinstance(val, ObjectId):
                        val = "\""+str(val)+"\""
                    else:
                        val = str(val).lower().replace("\"","\\\"")
                    phrase += val+" "
                else:
                    return False
            elif isinstance(children, list):
                res = self.evaluate(children, data)
                phrase += str(res).lower()+" "
            else:
                phrase += children.lower()+" "
        phrase = phrase.replace("false","False")
        phrase = phrase.replace("true","True")
        phrase = phrase.replace("none","None")
        return eval(phrase)

    def getIds(self, appTw):
        """Return Ids of the models in a PollenisatorTreeview matching the filter
        Args:
            appTw: a PollenisatorTreeview
        Raises:
            ParseError if Lark raises an UnexpectedToken or an UnexptectedCharacters exception.
        Returns:
            A list of Mongo iid which data matched the filter
        """
        found_res = []
        views = appTw.views
        datamanager = DataManager.getInstance()
        for dbId in views:
            view_object = views[dbId]["view"]
            data = view_object.controller.getData()
            if data is None:
                continue
            keys = list(data.keys())
            it_key = 0
            len_keys = len(keys)
            while it_key < len_keys :
                key = keys[it_key]
                val = data[key]
                if isinstance(val, dict):
                    for subkey, subval in val.items():
                        data[key+"."+str(subkey)] = subval
                        keys.append(key+"."+str(subkey))
                elif isinstance(val, list):
                    for i,subval in enumerate(val):
                        data[key+"["+str(i)+"]"] = subval
                it_key += 1
                len_keys = len(keys)
            dtype = view_object.controller.getType()
            data["type"] = dtype
            tag = datamanager.find("tags", {"item_id":ObjectId(dbId)}, multi=False)
            if tag is not None:
                data["tags"] = tag.tags
            else:
                data["tags"] = []
            research = []
            for tag in data["tags"]:
                tag = TagInfos(tag)
                tag_name = tag.name
                research.append(tag_name)
            data["tags"] = research
            result = self.evaluate(self.parsed, data)
            if result:
                found_res.append(dbId)
        return found_res



"""Show a message to the user for a limited amount of time
"""
import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
import threading


class ChildDialogToast(CTkToplevel):
    """floating basic window with info text inside, fading away after a short time.
    """

    def __init__(self, parent, text, *args, **kwargs):
        """
        create the floating window but do not display it

        Args:
            parent: the tkinter parent view to use for this window construction.
            text: the text to display inside the toast
            args: not used
            kwargs: 
                - if fadingTime is defined, delay before fading starts. Default to 1.0s
                - if x is defined, x position of the toast top left corner, else default to parent x +25 .
                - if y is defined, y position of the toast top left corner, else default to parent botoom - 100.
        """
        super().__init__()

        self.text = text
        self.parent = parent
        self.kwargs = kwargs

    def show(self):
        """
        show the floating window, fading on 1.0 sec
        """
        x = self.kwargs.get("x")
        if x is None:
            x = self.parent.winfo_rootx() + 25
        y = self.kwargs.get("y")
        if y is None:
            y = (self.parent.winfo_rooty() + self.parent.winfo_reqheight() - 100)
        # creates a toplevel window
        # Leaves only the label and removes the app window
        self.wm_overrideredirect(True)
        self.wm_geometry("+%d+%d" % (x, y))
        w = self.kwargs.get("width", 200)
        h = self.kwargs.get("height", 20)
        self.geometry("%dx%d" % (w, h))
        label = CTkLabel(self, text=self.text, justify='left')
        label.pack(fill=tk.BOTH)
        self.transparency = 1.0
        fadingTime = self.kwargs.get("fadingTime", 1.0)
        self.timer_fade = None
        if fadingTime is not None:
            self.timer_fade = threading.Timer(fadingTime, self.fade)
            self.timer_fade.start()
        try:
            self.wait_visibility()
            self.lift()
        except tk.TclError:
            pass
    
    def moveto(self, x, y):
        self.wm_geometry("+%d+%d" % (x, y))

    def fade(self):
        try:
            self.transparency -= 0.2
            if self.transparency <= 0.0:
                self.destroy()
                del self
                return
            self.wm_attributes("-alpha", self.transparency)
            self.timer_fade = threading.Timer(0.1, self.fade)
            self.timer_fade.start()
        except RuntimeError:
            return


import os
import traceback
from bson import ObjectId
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.models.metaelement import REGISTRY
from pollenisatorgui.core.components.logger_config import logger

class Subject:
    """Represents what is being observed"""
 
    def __init__(self):
        """create an empty observer list"""
        self._observers = []
 
    def notify(self, notif, obj, old_obj):
        """Alert the observers"""
        for observer in self._observers:
            observer.update_received(self, notif, obj, old_obj)
 
    def attach(self, observer):
        """If the observer is not in the list,
        append it into the list"""
        if type(observer) not in [type(o) for o in self._observers]:
            self._observers.append(observer)
 
    def detach(self, observer):
        """Remove the observer from the observer list"""
        try:
            self._observers.remove(observer)
        except ValueError:
            pass
 
class DataManager(Subject):
    __instances = dict()
    def __init__(self):
        super().__init__()
        self._observers
        self.data = {}
        self.ip_index = {}
        self.currentPentest = None
        pid = os.getpid()
        DataManager.__instances[pid] = self

    def openPentest(self, pentest):
        self.currentPentest = pentest
        self.data = {}


    # def load(self, collections=None, forceReload=False):
    #     if len(self.data) > 0 and not forceReload:
    #         return
    #     self.currentPentest = APIClient.getInstance().getCurrentPentest()
        
    #     for coll, model in REGISTRY.items():
    #         if collections is not None and coll not in collections:
    #             continue
    #         self.data[coll] = {}
    #         datas = model.fetchPentestObjects()
    #         for item in datas:
    #             self.data[coll][str(item.getId())] = item

    def get(self, collection, iid):
        collection = collection.lower()
        if collection not in REGISTRY.keys() and collection[:-1] in REGISTRY.keys():
            collection = collection[:-1]
        if iid == "*":
            ret = []
            datas = REGISTRY[collection].fetchPentestObjects()
            if self.data.get(collection, None) is None:
                self.data[collection] = {}
            for data in datas:
                self.data[collection][str(data._id)] = data
                if collection == "ip":
                    self.ip_index[data.ip] = data
                ret.append(data)
            return ret
        return self.find(collection, {"_id": iid}, multi=False, fetch_on_none=True)
        # collection = collection.lower()
        # if collection not in self.data.keys() and collection[:-1] in self.data.keys():
        #     collection = collection[:-1]
        # if collection not in self.data.keys():
        #     return default
        # if iid == "*":
        #     return self.data[collection]
        # return self.data[collection].get(str(iid), None)

    def find(self, collection, search, multi=True, fetch_on_none=True):
        collection = collection.lower()
        if collection not in REGISTRY.keys() and collection[:-1] in REGISTRY.keys():
            collection = collection[:-1]
        if collection not in self.data:
            self.data[collection] = {}
        if len(search) == 1:
            if "ip" in search.keys() and collection == "ip":
                ip = search["ip"]
                ret = self.ip_index.get(ip, None)
                if ret is None and fetch_on_none:
                    search = {"ip": ip}
                    ret = REGISTRY[collection].fetchObject(search)
                    self.data[collection][str(ret.getId())] = ret
                    self.ip_index[ip] = ret
                return ret
            if "_id" in search.keys():
                iid = search["_id"]
                ret = self.data[collection].get(str(iid), None)
                if ret is None and fetch_on_none:
                    search = {"_id": ObjectId(iid)}
                    if collection == "command":
                        ret = REGISTRY[collection].fetchObject(search, self.currentPentest)
                    else:
                        ret = REGISTRY[collection].fetchObject(search)
                    if ret is None:
                        return None
                    self.data[collection][str(ret.getId())] = ret
                return ret
        # if collection not in self.data.keys():
        #     return None
        ret = []
        for data_model in list(self.data.get(collection, {}).values()):
            is_match = True
            for key, val in search.items():
                data = data_model.getData()
                compared = data.get(key, None)
                if isinstance(compared, list):
                    if val not in compared:
                        is_match = False
                        break
                elif compared != val:
                    is_match = False
                    break
            if is_match:
                ret.append(data_model)
                if not multi:
                    return data_model
        if len(ret) == 0 and fetch_on_none:
            if multi:
                if collection == "command":
                        datas = REGISTRY[collection].fetchObjects(search, self.currentPentest)
                else:
                    datas = REGISTRY[collection].fetchObjects(search)
                for data in datas:
                    self.data[collection][str(data.getId())] = data
                    if collection == "ip":
                        self.ip_index[data.ip] = data
                    ret.append(data)
            else:
                if collection == "command":
                    data = REGISTRY[collection].fetchObject(search, self.currentPentest)
                else:
                    data = REGISTRY[collection].fetchObject(search)
                ret = data
        if not multi and not ret:
            return None
        return ret
            
    def getClass(self, class_str):
        for coll, model in REGISTRY.items():
            if coll == class_str or coll+"s" == class_str:
                return model
        raise ValueError("Class not found "+str(class_str))

    def remove(self, collection, iid):
        if collection not in self.data.keys() and collection+"s" in self.data.keys():
            collection = collection+"s"
        if collection not in self.data.keys() and collection[:-1] in self.data.keys():
            collection = collection[:-1]
        if collection not in self.data.keys():
            return
        try:
            del self.data[collection][str(iid)]
        except KeyError:
            pass
        
    def set(self, collection, iid, newVal):
        if collection not in self.data.keys() and collection+"s" in self.data.keys():
            collection = collection+"s"
        if collection not in self.data.keys() and collection[:-1] in self.data.keys():
            collection = collection[:-1]
        if collection not in self.data.keys():
            return 
        self.data[collection][str(iid)] = newVal
    
    @staticmethod
    def getInstance():
        """ Singleton Static access method.
        """
        pid = os.getpid()  # HACK : One api client per process.
        instance = DataManager.__instances.get(pid, None)
        if instance is None:
            DataManager()
        return DataManager.__instances[pid]

    def handleNotification(self, notification):
        try:
            apiclient = APIClient.getInstance()
            obj = None
            old_obj = None
            if notification["db"] != "pollenisator":
                class_name = notification["collection"]
                if class_name in self.data.keys() or class_name[:-1] in self.data.keys():
                    if notification["action"] == "update" or notification["action"] == "insert":
                        updated_data = apiclient.findInDb(notification["db"], notification["collection"], {"_id": ObjectId(notification["iid"])}, False, use_cache=False)
                        obj = self.getClass(class_name)(updated_data)
                        old_obj = self.get(class_name, notification["iid"])
                        if old_obj is None:
                            notification["action"] = "insert"
                        self.set(class_name, notification["iid"], obj)
                    elif notification["action"] == "insert_many" or notification["action"] == "update_many":
                        updated_data = apiclient.findInDb(notification["db"], notification["collection"], {"_id": {"$in":notification["iid"]}}, True)
                        obj = []
                        for data in updated_data:
                            try:
                                model = self.getClass(class_name)(data)
                            except Exception as e:
                                logger.error("Error while creating model : "+str(e))
                                continue
                            obj.append(model)
                            self.set(class_name, data["_id"], model)
                    elif notification["action"] == "delete":
                        self.remove(class_name, notification["iid"])
            self.notify(notification, obj, old_obj)
        except Exception as e:
            traceback.print_exc()
            logger.critical("Error while handling notification : "+str(e))
    
    

    


"""Scope Model"""

from pollenisatorgui.core.components.datamanager import DataManager
from pollenisatorgui.core.models.element import Element
from pollenisatorgui.core.models.ip import Ip
from bson.objectid import ObjectId
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.models.tool import Tool
import pollenisatorgui.core.components.utils as utils
from pollenisatorgui.core.components.settings import Settings


class Scope(Element):
    """
    Represents a Scope object that defines a scope that will be targeted by network or domain tools.

    Attributes:
        coll_name: collection name in pollenisator database
    """

    coll_name = "scopes"

    def __init__(self, valuesFromDb=None):
        """Constructor
        Args:
            valueFromDb: a dict holding values to load into the object. A mongo fetched interval is optimal.
                        possible keys with default values are : _id (None), parent (None), tags([]), infos({}),
                        wave(""), scope(""), notes("")
        """
        if valuesFromDb is None:
            valuesFromDb = {}
        super().__init__(valuesFromDb.get("_id", None), valuesFromDb.get("parent", None), valuesFromDb.get("infos", {}))
        self.initialize(valuesFromDb.get("wave", ""), valuesFromDb.get("scope", ""),
                        valuesFromDb.get("notes", ""), valuesFromDb.get("infos", {}))

    def initialize(self, wave, scope="", notes="", infos=None):
        """Set values of scope
        Args:
            wave: the wave parent of this scope
            scope: a string describing the perimeter of this scope (domain, IP, NetworkIP as IP/Mask)
            notes: notes concerning this IP (opt). Default to ""
            infos: a dictionnary of additional info
        Returns:
            this object
        """
        self.wave = wave
        self.scope = scope
        self.notes = notes
        self.infos = infos if infos is not None else {}
        return self

    def update(self, pipeline_set=None):
        """Update this object in database.
        Args:
            pipeline_set: (Opt.) A dictionnary with custom values. If None (default) use model attributes.
        """
        apiclient = APIClient.getInstance()
        if pipeline_set is None:
            apiclient.update("scopes", ObjectId(self._id), {"notes": self.notes})
        else:
            apiclient.update("scopes", ObjectId(self._id), pipeline_set)

    def delete(self):
        """
        Delete the Scope represented by this model in database.
        Also delete the tools associated with this scope
        Also remove this scope from ips in_scopes attributes
        """
        # deleting tool with scope lvl
        apiclient = APIClient.getInstance()
        apiclient.delete("scopes", ObjectId(self._id))
        # Finally delete the selected element

    def addInDb(self):
        """
        Add this scope in database.

        Returns: a tuple with :
                * bool for success
                * mongo ObjectId : already existing object if duplicate, create object id otherwise 
        """
        base = self.getDbKey()
        # Checking unicity
        apiclient = APIClient.getInstance()
        res_insert, iid = apiclient.insert("scopes", base)
        self._id = iid
        # adding the appropriate tools for this scope.
        return res_insert, iid

    def addDomainInDb(self, checkDomain=True):
        """
        Add this scope domain in database.

        Args:
            checkDomain: boolean. If true (Default), checks that the domain IP is in scope

        Returns: a tuple with :
                * bool for success
                * mongo ObjectId : already existing object if duplicate, create object id otherwise 
        """
        # Checking unicity
        base = self.getDbKey()
        apiclient = APIClient.getInstance()
        # Check if domain's ip fit in one of the Scope of the wave
        if checkDomain:
            if not Scope.checkDomainFit(self.wave, self.scope):
                return -1, None
        # insert the domains in the scopes
        res_insert, iid = apiclient.insert("scopes", base)
        self._id = iid
        # Adding appropriate tools for this scopes
        return 1, iid

    @classmethod
    def checkDomainFit(cls, waveName, domain):
        """
        Check if a found domain belongs to one of the scope of the given wave.

        Args:
            waveName: The wave id (name) you want to search for a validating scope
            domain: The found domain.

        Returns:
            boolean
        """
        # Checking settings for domain check.
        settings = Settings()
        # get the domain ip so we can search for it in ipv4 range scopes.
        domainIp = utils.performLookUp(domain)
        datamanager = DataManager.getInstance()
        scopesOfWave = datamanager.find("scopes", {"wave": waveName})
        for scopeOfWave in scopesOfWave:
            scopeIsANetworkIp = utils.isNetworkIp(scopeOfWave["scope"])
            if scopeIsANetworkIp:
                if settings.db_settings.get("include_domains_with_ip_in_scope", False):
                    if Ip.checkIpScope(scopeOfWave["scope"], domainIp):
                        return True
            else:  # If scope is domain
                # check if we include subdomains
                if settings.db_settings.get("include_all_domains", False):
                    return True
                else:
                    splitted_domain = domain.split(".")
                    # Assuring to check only if there is a domain before the tld (.com, .fr ... )
                    topDomainExists = len(splitted_domain) > 2
                    if topDomainExists:
                        if settings.db_settings.get("include_domains_with_topdomain_in_scope", False):
                            if splitted_domain[1:] == scopeOfWave["scope"].split("."):
                                return True
                    if settings.db_settings.get("include_domains_with_ip_in_scope", False):
                        inRangeDomainIp = utils.performLookUp(
                            scopeOfWave["scope"])
                        if str(inRangeDomainIp) == str(domainIp):
                            return True
        return False

    

    def _getParentId(self):
        """
        Return the mongo ObjectId _id of the first parent of this object. For a scope it is the wave.

        Returns:
            Returns the parent wave's ObjectId _id".
        """
        datamanager = DataManager.getInstance()
        res = datamanager.find("waves", {"wave": self.wave}, False)
        if res is None:
            return None
        return res.getId()

    def __str__(self):
        """
        Get a string representation of a scope.

        Returns:
            Returns the scope string (network ipv4 range or domain).
        """
        return self.scope

    def getData(self):
        """Return scope attributes as a dictionnary matching Mongo stored scopes
        Returns:
            dict with keys wave, scope, notes, _id, tags and infos
        """
        return {"wave": self.wave, "scope": self.scope, "notes": self.notes, "_id": self.getId(), "infos": self.infos}

    def getDbKey(self):
        """Return a dict from model to use as unique composed key.
        Returns:
            A dict (2 keys :"wave", "scope")
        """
        return {"wave": self.wave, "scope": self.scope}

    

    def isDomain(self):
        """Returns True if this scope is not a valid NetworkIP
        Returns:
            bool
        """
        return not utils.isNetworkIp(self.scope)

    @classmethod
    def isSubDomain(cls, parentDomain, subDomainTest):
        """Returns True if this scope is a valid subdomain of the given domain
        Args:
            parentDomain: a domain that could be the parent domain of the second arg
            subDomainTest: a domain to be tested as a subdomain of first arg
        Returns:
            bool
        """
        splitted_domain = subDomainTest.split(".")
        # Assuring to check only if there is a domain before the tld (.com, .fr ... )
        topDomainExists = len(splitted_domain) > 2
        if topDomainExists:
            if ".".join(splitted_domain[1:]) == parentDomain:
                return True
        return False
